
import os
import json


from epyk.core.js import Imports

from epyk.core.js import JsUtils
from epyk.core.js import JsNavigator
from epyk.core.js import JsWindow
from epyk.core.js import JsLocation
from epyk.core.js import JsMaths
from epyk.core.js import JsPerformance

# All the predefined Javascript Statements
from epyk.core.js.statements import JsIf
from epyk.core.js.statements import JsFor
from epyk.core.js.statements import JsWhile
from epyk.core.js.statements import JsSwitch

# All the predefined variable types
from epyk.core.js.fncs import JsFncs
from epyk.core.js.primitives import JsArray
from epyk.core.js.primitives import JsObjects
from epyk.core.js.primitives import JsObject
from epyk.core.js.primitives import JsNumber
from epyk.core.js.primitives import JsString

from epyk.core.js.objects import JsNodeAttributes
from epyk.core.js.objects import JsNodeDom
from epyk.core.js.objects import JsData


class JsConsole(object):
  """
  Description:
  ------------
  This is a wrapper to the Console

  Related Pages:

    https://medium.freecodecamp.org/how-to-get-the-most-out-of-the-javascript-console-b57ca9db3e6d
  """

  @property
  def debugger(self):
    """
    Description:
    ------------
    Trigger a Javascript debugger from this point.
    The Javascript will be stopped and it will be possible to check the process step by step in the browser using F12

    Usage::

      rptObj.js.console.debugger

    Related Pages:

			https://www.w3schools.com/jsref/jsref_debugger.asp

    :return: The Javascript Keyword to trigger the browser debugger
    """
    return JsObject.JsKeyword("debugger")

  @property
  def clear(self):
    """
    Description:
    ------------
    The console.clear() method clears the console.

    Usage::

      rptObj.js.console.clear

    Related Pages:

			https://www.w3schools.com/jsref/met_console_clear.asp

    :return: The Javascript String used to clear the console (F12 in standard browsers)
    """
    return JsFncs.JsFunction("console.clear()")

  def log(self, jsData, jsFnc=None, skip_data_convert=False):
    """
    Description:
    ------------
    The console.log() method writes a message to the console.

    Usage::

      rptObj.js.console.log("Test")

    Related Pages:

			https://www.w3schools.com/jsref/met_console_log.asp

    Attributes:
    ----------
    :param jsData: The Javascript fragment
    :param jsFnc:
    :param skip_data_convert:

    :return: The Javascript String used to clear the console (F12 in standard browsers)
    """
    if skip_data_convert:
      return JsFncs.JsFunction("console.log(%s)" % jsData)

    return JsFncs.JsFunction("console.log(%s)" % JsUtils.jsConvertData(jsData, jsFnc))

  def info(self, jsData, jsFnc=None):
    """
    Description:
    ------------
    The console.info() method writes a message to the console.

    Related Pages:

			https://www.w3schools.com/jsref/met_console_info.asp

    Attributes:
    ----------
    :param data: The Javascript fragment
    :param jsFnc:

    :return: The Javascript String used to clear the console (F12 in standard browsers)
    """
    return JsFncs.JsFunction("console.info(%s)" % JsUtils.jsConvertData(jsData, jsFnc))

  def warn(self, jsData, jsFnc=None):
    """
    Description:
    ------------
    The console.warn() method writes a warning to the console.

    Related Pages:

			https://www.w3schools.com/jsref/met_console_warn.asp

    Attributes:
    ----------
    :param data: The Javascript fragment
    :param jsFnc:

    :return: The Javascript String used to clear the console (F12 in standard browsers)
    """
    return JsFncs.JsFunction("console.warn(%s)" % JsUtils.jsConvertData(jsData, jsFnc))

  def error(self, jsData, jsFnc=None):
    """
    Description:
    ------------
    The console.error() method writes an error message to the console.

    Related Pages:

			https://www.w3schools.com/jsref/met_console_error.asp

    Attributes:
    ----------
    :param data: The Javascript fragment
    :param isPyData:

    :return: The Javascript String used to clear the console (F12 in standard browsers)
    """
    return JsFncs.JsFunction("console.error(%s)" % JsUtils.jsConvertData(jsData, jsFnc))

  def table(self, jsData, jsHeader=None):
    """
    Description:
    ------------
    The console.table() method writes a table in the console view.

    Related Pages:

			https://www.w3schools.com/jsref/met_console_table.asp

    Attributes:
    ----------
    :param jsData: Required. The data to fill the table with
    :param jsHeader: Optional. An array containing the names of the columns to be included in the table
    :return: The Javascript String used to clear the console (F12 in standard browsers)
    """
    if jsHeader is not None:
      return JsFncs.JsFunction("console.table(%s, %s)" % (jsData, jsHeader))

    return JsFncs.JsFunction("console.table(%s)" % jsData)

  def time(self, htmlId):
    """
    Description:
    ------------
    The console.time() method starts a timer in the console view

    Related Pages:

			https://www.w3schools.com/jsref/met_console_time.asp

    Attributes:
    ----------
    :param htmlId: Use the label parameter to give the timer a name
    :return: A Python Javascript Number
    """
    return JsNumber.JsNumber("console.time('%s')" % htmlId, isPyData=False)

  def timeEnd(self, htmlId):
    """
    Description:
    ------------
    The console.timeEnd() method ends a timer, and writes the result in the console view.

    Related Pages:

			https://www.w3schools.com/jsref/met_console_timeend.asp

    Attributes:
    ----------
    :param htmlId: The name of the timer to end
    :return: The Javascript String used to clear the console (F12 in standard browsers)
    """
    return JsFncs.JsFunction("console.timeEnd('%s')" % htmlId)

  def _assert(self, jsData, strInfo, jsFnc=None):
    """
    Description:
    ------------
    The console.assert() method writes a message to the console, but only if an expression evaluates to false.

    Related Pages:

			https://www.w3schools.com/jsref/met_console_assert.asp

    Attributes:
    ----------
    :param jsData:
    :param strInfo:
    :param isPyData:
    """
    return JsFncs.JsFunction("console.assert('%s', '%s')" % (JsUtils.jsConvertData(jsData, jsFnc), strInfo))

  def tryCatch(self, jsFncs, jsFncsErrs="console.warn(err.message)"):
    """
    Description:
    ------------
    Javascript Try Catch Exceptions

    Related Pages:

			https://www.w3schools.com/jsref/jsref_obj_error.asp

    Attributes:
    ----------
    :param jsFncs:
    :param jsFncsErrs:

    :return: The Javascript String used to clear the console (F12 in standard browsers)
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    return JsFncs.JsFunction("try{%s} catch(err){%s}" % (";".join(jsFncs), jsFncsErrs))


class JsJson(object):
  """
  Description:
  ------------
  Wrapper around the Javascript Json module
  This wrapper will only wrapper the different functions available in the underlying library.
  The documentation can be found in each function or are available on the Javascript Official documentation

  Related Pages:

    https://www.w3schools.com/js/js_json_intro.asp
  """

  def parse(self, jsData, jsResultFnc=None, jsFnc=None):
    """
    Description:
    ------------
    Parses a JSON string and returns a JavaScript object

    Related Pages:

			https://www.w3schools.com/js/js_json_parse.asp
      https://www.w3schools.com/jsref/jsref_parse_json.asp

    Attributes:
    ----------
    :param jsData: Required. A string written in JSON format
    :param jsResultFnc: Optional. A function used to transform the result. The function is called for each item. Any nested objects are transformed before the parent.

    :return: The Javascript string method
    """
    jsData = JsUtils.jsConvertData(jsData, jsFnc)
    if jsResultFnc is not None:
      return JsFncs.JsFunction("JSON.parse(%s, %s)" % (jsData, jsResultFnc))

    return JsFncs.JsFunction("JSON.parse(%s)" % jsData)

  def stringify(self, jsData, replacer=None, space=0, jsFnc=None):
    """
    Description:
    ------------
    The JSON.stringify() method converts JavaScript objects into strings.

    Related Pages:

			https://www.w3schools.com/js/js_json_stringify.asp

    Attributes:
    ----------
    :param jsData: Required. The value to convert to a string
    :param replacer: Optional. Either a function or an array used to transform the result. The replacer is called for each item.
    :param space: Optional. Either a String or a Number. A string to be used as white space (max 10 characters),
      or a Number, from 0 to 10, to indicate how many space characters to use as white space.

    :return: The Javascript string method
    """
    return JsString.JsString("JSON.stringify(%s, %s, %s)" % (JsUtils.jsConvertData(jsData, jsFnc), json.dumps(replacer), space), isPyData=False)


class JsBreadCrumb(object):

  def __init__(self, src=None):
    self._src = src
    self._selector = "breadcrumb"
    self._anchor = None
    self._src._props.setdefault('js', {}).setdefault('builders', []).append("%s = {pmts: %s}" % (self._selector, json.dumps(self._src.http)))

  def add(self, key, jsData):
    """
    Description:
    ------------
    Add an entry to the Javascript breadcrumb dictionary

    Attributes:
    ----------
    :param key: The key in the Breadcrumb dictionary
    :param jsData:

    :return: Nothing
    """
    return JsFncs.JsFunction('%s["pmts"]["%s"] = %s' % (self._selector, key, JsUtils.jsConvertData(jsData, None)))

  def get(self, key=None):
    """
    Description:
    ------------
    returns the object stored in the breadcrumb dictionary

    Attributes:
    ----------
    :param key: Optinal. The key in the Breadcrumb dictionary

    :return: A Python object
    """
    if key is None:
      return JsObject.JsObject("%s" % self._selector)

    return JsObject.JsObject('%s["pmts"]["%s"]' % (self._selector, key))

  def hash(self, jsData):
    """
    Description:
    ------------
    Add an anchor to the URL after the hash tag

    Related Pages:

			https://www.w3schools.com/jsref/prop_loc_hash.asp
    """
    return JsObject.JsObject('%s["anchor"] = %s' % (self._selector, JsUtils.jsConvertData(jsData, None)))

  @property
  def url(self):
    """
    Description:
    ------------
    Get the full URL
    """
    js_location = JsLocation.JsLocation()
    origin = js_location.origin
    pathname = js_location.pathname
    return JsString.JsString(origin + pathname + "?" + JsObject.JsObject(self.toStr()))

  def toStr(self):
    return '%s(%s)' % (JsFncs.FncOnRecords(self._src._props['js']).url(), self._selector)


class JsScreen(object):

  @property
  def availHeight(self):
    """
    Description:
    ------------
    The availHeight property returns the height of the user's screen, in pixels, minus interface features like the Windows Taskbar.

    Related Pages:

			https://www.w3schools.com/jsref/prop_screen_availheight.asp
    """
    return JsNumber.JsNumber("screen.availHeight")

  @property
  def availWidth(self):
    """
    Description:
    ------------
    The availWidth property returns the width of the user's screen, in pixels, minus interface features like the Windows Taskbar.

    Related Pages:

			https://www.w3schools.com/jsref/prop_screen_availwidth.asp
    """
    return JsNumber.JsNumber("screen.availWidth")

  @property
  def colorDepth(self):
    """
    Description:
    ------------
    The colorDepth property returns the bit depth of the color palette for displaying images (in bits per pixel).

    Related Pages:

			https://www.w3schools.com/jsref/prop_screen_colordepth.asp
    """
    return JsNumber.JsNumber("screen.colorDepth")

  @property
  def height(self):
    """
    Description:
    ------------
    The height property returns the total height of the user's screen, in pixels.

    Related Pages:

			https://www.w3schools.com/jsref/prop_screen_height.asp
    """
    return JsNumber.JsNumber("screen.height")

  @property
  def pixelDepth(self):
    """
    Description:
    ------------
    The pixelDepth property returns the color resolution (in bits per pixel) of the visitor's screen.

    Related Pages:

			https://www.w3schools.com/jsref/prop_screen_pixeldepth.asp
    """
    return JsNumber.JsNumber("screen.pixelDepth")

  @property
  def width(self):
    """
    Description:
    ------------
    The width property returns the total width of the user's screen, in pixels.

    Related Pages:

			https://www.w3schools.com/jsref/prop_screen_width.asp
    """
    return JsNumber.JsNumber("screen.width")


class JsBase(object):
  class __internal(object):
    _props, _context, jsOnLoadEvtsFnc, http = {}, {}, [], []
    jsImports, cssImport = set([]), set([])

  def __init__(self, src=None):
    self._src = src if src else self.__internal() # The underlying source object is not supposed to be touched in the underlying classes
    self.console = JsConsole()
    self.localStorage = JsWindow.JsLocalStorage()
    self.window = JsWindow.JsWindow(self)
    self.performance = JsPerformance.JsPerformance(self)
    self.sessionStorage = JsWindow.JsSessionStorage()
    self.json = JsJson()
    self.math = JsMaths.JsMaths()

    # shortcut functions
    self.alert = self.window.alert
    self.log = self.console.log
    self._breadcrumb, self.__data, self.__location = None, None, None

  @property
  def viewHeight(self):
    """
    Description:
    -----------
    Return the current View port height visible in the browser
    """
    return JsNumber.JsNumber("Math.max(%s, %s)" % (self.documentElement.clientHeight, self.window.innerHeight))

  @property
  def documentElement(self):
    """
    Description:
    -----------
    Document.documentElement returns the Element that is the root element of the document (for example, the <html> element for HTML documents).

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/Document/documentElement
    """
    return JsNodeDom.JsDoms.get("document.documentElement")

  @property
  def screen(self):
    """
    Description:
    ------------
    The screen object contains information about the visitor's screen.

    Related Pages:

			https://www.w3schools.com/jsref/obj_screen.asp
    """
    return JsScreen()

  @property
  def navigator(self):
    """
    Description:
    ------------
    The information from the navigator object can often be misleading, and should not be used to detect browser versions because:

      - Different browsers can use the same name
      - The navigator data can be changed by the browser owner
      - Some browsers misidentify themselves to bypass site tests
      - Browsers cannot report new operating systems, released later than the browser
    """
    return JsNavigator.JsNavigator(self)

  @property
  def location(self):
    """
    Description:
    ------------
    Property to the Javascript Location functions

    Related Pages:

			https://www.w3schools.com/jsref/obj_location.asp

    :rtype: JsLocation.JsLocation
    """
    if self.__location is None:
      self.__location = JsLocation.JsLocation()
    return self.__location

  @property
  def objects(self):
    """
    Description:
    ------------
    Interface to the main Javascript Classes and Primitives

    :rtypw: JsObjects.JsObjects
    """
    return JsObjects.JsObjects(self)

  @property
  def jquery(self):
    """
    Description:
    ------------
    jQuery is a fast, small, and feature-rich JavaScript library.
    It makes things like HTML document traversal and manipulation, event handling, animation, and Ajax much simpler with an easy-to-use API that works across a multitude of browsers.
    With a combination of versatility and extensibility, jQuery has changed the way that millions of people write JavaScript.

    Related Pages:

			https://jquery.com/
    """
    from epyk.core.js.packages import JsQuery

    return JsQuery.JQuery(None)

  @property
  def d3(self):
    """
    Description:
    ------------
    D3.js is a JavaScript library for manipulating documents based on data.
    D3 helps you bring data to life using HTML, SVG, and CSS.
    D3’s emphasis on web standards gives you the full capabilities of modern browsers without tying yourself to a proprietary framework, combining powerful visualization components and a data-driven approach to DOM manipulation.

    Related Pages:

			https://d3js.org/
    """
    from epyk.core.js.packages import JsD3

    return JsD3.JsD3(self._src, "d3")

  def not_(self, data):
    """
    Description:
    ------------
    Add the Symbol (!) for the boolean negation.
    This feature is also available directly to any JsBoolean objects

    Usage::

      jsObj.not_(jsObj.objects.boolean.get("weekend"))

    Related Pages:

			https://developer.mozilla.org/fr/docs/Web/JavaScript/Reference/Op%C3%A9rateurs/Op%C3%A9rateurs_logiques

    Attributes:
    ----------
    :param data: The Javascript Object considered as a boolean

    :return: The Javascript fragment string
    """
    return JsFncs.JsFunction("!%s" % JsUtils.jsConvertData(data, None))

  def if_(self, jsCond, jsFnc):
    """
    Description:
    ------------
    Conditional statements are used to perform different actions based on different conditions.

    Related Pages:

			https://www.w3schools.com/js/js_if_else.asp

    Attributes:
    ----------
    :param jsCond:
    :param jsFnc:
    """
    if isinstance(jsCond, list):
      jsCond = "(%s)" % ")||(".join(JsUtils.jsConvertFncs(jsCond))
    self.__if = JsIf.JsIf(jsCond, jsFnc, self._src)
    return self.__if

  def return_(self, jsData):
    """
    Description:
    ------------
    Javascript return keyword

    Related Pages:
:param jsData: The Javascript expression
    """
    return JsFncs.JsFunction("return %s" % jsData)

  def switch(self, jsObj):
    """
    Description:
    ------------

    Related Pages:

			https://www.w3schools.com/js/js_switch.asp

    Attributes:
    ----------
    :param jsFnc:
    """
    if not hasattr(jsObj, 'varName'):
      if isinstance(jsObj, list):
        jsObj = JsArray.JsArray(jsObj, setVar=True)
      else:
        jsObj = JsObject.JsObject(jsObj, setVar=True)
    self.__switch = JsSwitch.JsSwitch(jsObj, self._context)
    return self.__switch

  def while_(self, pivot, jsFnc=None, iterVar='i', start=0, step=1):
    """
    Description:
    ------------

    :param jsCond:
    """
    jsPivot = JsUtils.jsConvertData(pivot, jsFnc)
    if isPyData and isinstance(pivot, (list, range)):
      self.__while = JsWhile.JsWhile(jsPivot, ruleType='array', iterVar=iterVar, start=start, step=step, context=self._context)
    elif isPyData and isinstance(pivot, dict):
      self.__while = JsWhile.JsWhile(jsPivot, ruleType='dict', iterVar=iterVar, start=start, step=step, context=self._context)
    else:
      self.__while = JsWhile.JsWhile(jsPivot, self._context)
    return self.__while

  def for_(self, iterable, jsDataKey=None, isPyData=False, jsFnc=None, iterVar='i', start=0, step=1):
    """
    Description:
    ------------

    :param iterable:
    """
    jsIterable = JsUtils.jsConvert(iterable, jsDataKey, isPyData, jsFnc)
    self.__for = JsFor.JsFor(jsIterable, iterVar, start, step, self._context)
    return self.__for

  def custom(self, jsData, jsDataKey=None, isPyData=False, jsFnc=None):
    """
    Description:
    ------------
    Allow the definition of bespoke javascript strings
    """
    jsData = JsUtils.jsConvert(jsData, jsDataKey, isPyData, jsFnc)
    self._src._props.setdefault('js', {}).setdefault('bespoke', []).append(jsData)

  def customText(self, text):
    """
    Description:
    ------------
    Javascript fragment added at the begining of the page.
    This will be called before any function in the framework

    Attributes:
    ----------
    :param text: String. The Javascript fragment

    :return: self to allow the chaning
    """
    self._src._jsText.append(text)
    return self

  def customFile(self, filename, path=None):
    """
    Description:
    ------------
    This will load your local javascript file when the report will be built.
    Then you will be able to use the new features in the different Javascript wrappers

    Attributes:
    ----------
    :param filename: String. The file name
    :param path: String. optional. The file path

    :return:
    """
    if path is None:
      self._src.jsLocalImports.add("%s/js/%s" % (Imports.STATIC_PATH.replace("\\", "/"), filename))
    else:
      self._src.jsLocalImports.add("%s/%s" % (path, filename))
    return self

  def _addImport(self, importAlias):
    """
    Description:
    ------------

    :param importAlias:
    """
    self._src._props.setdefault('js', {}).setdefault('imports', set([])).add(importAlias)
    return self

  def extendProto(self, pyClass, fncName, jsFncs, pmts=None):
    """
    Description:
    ------------
    Javascript Framework extension

    Hook in the base class to allow the definition of specific function to add extra primitive features.
    Usual this function should be used in a wrapper function with the same name in order to have a coherent bridge between
    Python and Javascript.

    Related Pages:

			https://www.w3schools.com/js/js_object_prototypes.asp

    Attributes:
    ----------
    :param pyClass: The PyJs Classname
    :param fncName: The Javascript function name
    :param jsFncs: The Javascript function definition
    :param pmts: The Javascript function parameters

    :return: The JsObject
    """
    jsData = JsUtils.jsConvertFncs(jsFncs)
    self._src._props.setdefault('js', {}).setdefault('prototypes', {})["%s.prototype.%s" % (pyClass._jsClass, fncName)] = {"content": ";".join(jsData), 'pmts': pmts}
    return self

  def request_http(self, varName, method_type, url):
    """
    Description:
    ------------
    All modern browsers have a built-in XMLHttpRequest object to request data from a server.

    Related Pages:

			https://www.w3schools.com/xml/xml_http.asp

    Usage::

      rptObj.js.request_http("ajax", "POST", "https://api.cdnjs.com/libraries").setHeaders(header).onSuccess([
      rptObj.js.alert(rptObj.js.objects.request.get("ajax").responseText)]).send(encodeURIData={"search": 'ractive'})

    Attributes:
    ----------
    :param varName: String. The variable name created in the Javascript
    :param method_type: String. The method of the HTTP Request
    :param url: String. The url path of the HTTP request

    :rtype: JsObjects.XMLHttpRequest
    """
    method_type = JsUtils.jsConvertData(method_type, None)
    url = JsUtils.jsConvertData(url, None)
    return JsObjects.XMLHttpRequest(self._src, varName, method_type, url)

  def request_rpc(self, varName, method_type, fnc, url, extra_params=None):
    """
    Description:
    ------------
    Internal RPC to trigger services

    Attributes:
    ----------
    :param varName: String. The variable name created in the Javascript
    :param method_type: String. The method type
    :param fnc: Fnc. A python function
    :param url: String. The service url

    :rtype: JsObjects.XMLHttpRequest
    """
    if not extra_params:
      extra_params = {}
    extra_params = JsUtils.jsConvertData(extra_params, None)
    method_type = JsUtils.jsConvertData(method_type, None)
    url = JsUtils.jsConvertData(url, None)
    mod_name = fnc.__module__
    if mod_name == "__main__":
      import __main__

      mod_path, mod_name = os.path.split(__main__.__file__[:-3])
    else:
      mod_path = os.path.abspath(os.path.dirname(fnc.__module__))
    rpc_params = {"function": fnc.__name__, 'module': mod_name, 'path': mod_path, 'extra_params': extra_params}
    return JsObjects.XMLHttpRequest(self._src, varName, method_type, url, rpc_params)

  @property
  def fncs(self):
    """
    Description:
    ------------
    Property to the predefined Javascript functions

    :rtype: JsFncs.JsRegisteredFunctions

    :return: The predefined functions
    """
    return JsFncs.JsRegisteredFunctions(self._src)

  @property
  def breadcrumb(self):
    """
    Description:
    ------------
    Create a internal Breadcrumb to keep track of the user journey within your page.

    Related Pages:

			https://www.w3schools.com/howto/howto_css_breadcrumbs.asp

    :rtype: JsBreadCrumb

    :return: A Python breadcumb object
    """
    if self._breadcrumb is None:
      self._breadcrumb = JsBreadCrumb(self._src)
    return self._breadcrumb

  def navigateTo(self, url, options=None):
    """
    Description:
    ------------
    Navigator to another URL like NodeJs

    Usage::

      icon.click([self.context.rptObj.js.navigateTo(url)])

    Related Pages:

			https://redfin.github.io/react-server/annotated-src/navigateTo.html

    Attributes:
    ----------
    :param url:
    :param options:
    """
    return self.location.open_new_tab(url=url)

  def registerFunction(self, fncName, jsFncs, pmts=None):
    """
    Description:
    ------------
    Javascript Framework extension

    Register a predefined Javascript function
    This is only dedicated to specific Javascript transformation functions

    Description:
    ------------
    :param fncName: The function name
    :param jsFncs: Optional. The Javascript function definition
    :param pmts: Optional

    :return: The JsObject
    """
    jsData = JsUtils.jsConvertFncs(jsFncs)
    self._src._props.setdefault('js', {}).setdefault('functions', {})[fncName] = {'content': ";".join(jsData), 'pmt': pmts}
    return self

  def clipboard(self, jsData):
    """
    Description:
    ------------
    Copy the full URL to rhe clipboard

    Related Pages:

			https://isabelcastillo.com/hidden-input-javascript
    """
    return JsFncs.JsFunction('''
        var elInput = document.createElement('input'); 
        elInput.setAttribute('type', 'text');
        elInput.setAttribute('value', %s); document.body.appendChild(elInput);
        document.execCommand('copy', false, elInput.select()); elInput.remove()''' % JsUtils.jsConvertData(jsData, None))

  def addOnLoad(self, jsFncs):
    """
    Description:
    ------------
    Add javascript fragment dedicated to build the HTML page during the onLoad event.

    Attributes:
    ----------
    :param jsFncs: The Javascript functions to be added to this section
    """
    self._src._props.setdefault('js', {}).setdefault('builders', []).append(";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def addKeyEvent(self, jsFncs, keyCode=None, keyCondition=None, event="onkeyup"):
    """
    Description:
    ------------
    Add keyboard event to the document

    Either the keyCode or the keyCondition can be None

    Related Pages:
http://gcctech.org/csc/javascript/javascript_keycodes.htm

    Attributes:
    ----------
    :param jsFncs: The Javascript function
    :param keyCode: Optional. The keycode as an integer
    :param keyCondition: Optional. A special condition based on the code
    """
    if keyCode is None and keyCondition is None:
      raise Exception("keyCode or keyCondition must be defined")

    if keyCode is not None and keyCondition is not None:
      raise Exception("keyCode or keyCondition cannot be both defined")

    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]

    if keyCode is not None:
      keyCondition = "e.which == %s" % keyCode
    self._src._props.setdefault('js', {}).setdefault('builders', []).append("document.%s = function(e){if(%s) {%s}}" % (event, keyCondition, ";".join(JsUtils.jsConvertFncs(jsFncs))))
    return self

  def preload(self, ajaxParams):
    """
    Description:
    ------------
    Preload feature to be able to produce pre cached files.
    Those files will be produced in a asynchrone way and they will facilitate the fluidity in the web dashboard.
    The success file is only used as an indicator to check if the preload function has to be started. If the file is already present it will not be triggered.
    The status will be based on the return of the service in the query return (true / false

    Tip: Put a variable result to your service return in order to change the icon according to the status of your asynchronous call

    Usage::

      report.preload([{'url': "Test.py", 'success': 'test.csv'}])

    Attributes:
    ----------
    :param ajaxParams:

    :return: The Python object itself
    """
    pass

  def addOnReady(self, jsFncs):
    """
    Description:
    ------------
    The ready event occurs when the body DOM (document object model) has been loaded.

    Related Pages:

			https://www.w3schools.com/jquery/event_ready.asp

    Attributes:
    ----------
    :param jsFncs: The Javascript functions to be added to this section
    """
    self._src._props['js']['onReady'].add(";".join(JsUtils.jsConvertFncs(jsFncs)))
    return self

  def profile(self, type, htmlId, mark, recordsCount=""):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param type:
    :param htmlId:
    :param mark:
    :param recordsCount:
    """
    return "profileObj.push({type: '%s', htmlId: '%s', mark: '%s', records: %s})" % (type, htmlId, mark, recordsCount)


  # -----------------------------------------------------------------
  #                         DOCUMENTATION FUNCTIONS
  #
  # https://www.w3schools.com/jsref/dom_obj_document.a
  def getElementById(self, idName):
    """
    Description:
    ------------
    The getElementById() method returns the element that has the ID attribute with the specified value.

    Related Pages:

			https://www.w3schools.com/jsref/met_document_getelementbyid.asp

    Attributes:
    ----------
    :param idName: Required. The ID attribute's value of the element you want to get

    :return: An Element Object, representing an element with the specified ID. Returns null if no elements with the specified ID exists
    """
    return JsNodeDom.JsDoms("document.getElementById('%s')" % idName)

  def getElementsByName(self, name):
    """
    Description:
    ------------
    The getElementsByName() method returns a collection of all elements in the document with the specified name (the value of the name attribute), as a NodeList object.

    The NodeList object represents a collection of nodes. The nodes can be accessed by index numbers. The index starts at 0.

    Related Pages:

			https://www.w3schools.com/jsref/met_doc_getelementsbyname.asp

    Attributes:
    ----------
    :param name: Required. The name attribute value of the element you want to access/manipulate

    :return: A NodeList object, representing a collection of elements with the specified name.
             The elements in the returned collection are sorted as they appear in the source code.
    :rtype: JsNodeDom.JsDomsList
    """
    return JsNodeDom.JsDomsList.get(varName="document.getElementsByName(%s)" % JsUtils.jsConvertData(name, None))

  def getElementsByTagName(self, tagName, i=0):
    """
    Description:
    ------------
    The getElementsByTagName() method returns a collection of an elements's child elements with the specified tag name, as a NodeList object.

    The NodeList object represents a collection of nodes. The nodes can be accessed by index numbers. The index starts at 0.

    Related Pages:

			https://www.w3schools.com/jsref/met_element_getelementsbytagname.asp

    Attributes:
    ----------
    :param tagName: Required. The tagname of the child elements you want to get
    :param i:
    """
    return JsNodeDom.JsDoms("document.getElementsByTagName('%s')[%s]" % (tagName, i), varName="%s_%s" % (tagName, i), setVar=True)

  def createElement(self, tagName, varName=None, setVar=True, dom_id=None):
    """
    Description:
    ------------
    The createElement() method creates an Element Node with the specified name.

    Related Pages:

			https://www.w3schools.com/jsref/met_document_createelement.asp

    Attributes:
    ----------
    :param tagName: Required. The name of the element you want to create
    :param varName: Optional. The variable name to be set. Default random name
    :param setVar: Optional. Create a variable for the new object. Default True
    :param dom_id:
    """
    dom_obj = JsNodeDom.JsDoms.new(tagName, varName=varName, setVar=setVar, report=self._src)
    if dom_id is not None:
      dom_obj.attr("id", dom_id)
    return dom_obj

  def createTextNode(self, jsString=None):
    """
    Description:
    ------------
    The createTextNode() method creates a Text Node with the specified text.

    Related Pages:

			https://www.w3schools.com/jsref/met_document_createtextnode.asp

    Attributes:
    ----------
    :param jsString: Required. The text of the Text node

    :return: A Text Node object with the created Text Node
    """
    return JsObject.JsObject("document.createTextNode(%s)" % JsUtils.jsConvertData(jsString, None), isPyData=False)

  def encodeURIComponent(self, uri):
    """
    Description:
    ------------
    The encodeURIComponent() function encodes a URI component

    Related Pages:

			https://www.w3schools.com/jsref/jsref_encodeuricomponent.asp

    Attributes:
    ----------
    :param uri: Required. The URI to be encoded

    :return: A String, representing the encoded URI
    """
    return JsObject.JsObject("encodeURIComponent(%s)" % JsUtils.jsConvertData(uri, None))

  def decodeURIComponent(self, url_enc):
    """
    Description:
    ------------
    The decodeURIComponent() function decodes a URI component.

    Related Pages:

			https://www.w3schools.com/jsref/jsref_decodeuricomponent.asp

    Attributes:
    ----------
    :param url_enc: Required. The URI to be decoded

    :return: A String, representing the decoded URI
    """
    return JsObject.JsObject("decodeURIComponent(%s)" % JsUtils.jsConvertData(url_enc, None))

  @property
  def body(self):
    return JsNodeDom.JsDoms("document.body", setVar=False, isPyData=False)

  @property
  def data(self):
    """

    :rtype: JsData.JsData
    """
    if self.__data is None:
      self.__data = JsData.JsData(self._src)
    return self.__data

  def string(self, data, varName=None, setVar=False, isPyData=True):
    """
    Description:
    ------------
    Shortcut to the Javascript String primitives

    Attributes:
    ----------
    :param data:
    :param varName:
    :param setVar:
    :param isPyData:
    """
    return JsString.JsString(data, varName, setVar, isPyData, report=self._src)

  def number(self, data, varName=None, setVar=False, isPyData=True):
    """
    Description:
    ------------
    Shortcut to the Javascript Number primitives

    Attributes:
    ----------
    :param data:
    :param varName:
    :param setVar:
    :param isPyData:
    """
    return JsNumber.JsNumber(data, varName, setVar, isPyData, report=self._src)

  def object(self, data, varName=None, setVar=False, isPyData=True):
    """
    Description:
    ------------
    Shortcut to the Javascript Object primitives

    Attributes:
    ----------
    :param data:
    :param varName:
    :param setVar:
    :param isPyData:
    """
    return JsObject.JsObject(data, varName, setVar, isPyData, report=self._src)

  def activeElement(self):
    """
    Description:
    ------------
    The activeElement property returns the currently focused element in the document.

    Related Pages:

			https://www.w3schools.com/jsref/prop_document_activeelement.asp

    :return: A reference to the element object in the document that has focus
    """
    return JsNodeDom.JsDoms("document.activeElement")

  def title(self, text=None):
    """
    Description:
    ------------
    The title property sets or returns the title of the current document (the text inside the HTML title element).

    Related Pages:

			https://www.w3schools.com/jsref/prop_doc_title.asp

    Attributes:
    ----------
    :param text: A String, representing the title of the document
    """
    if text is None:
      return JsString.JsString(text)

    return "document.title = %s" % JsUtils.jsConvertData(text, None)

  def getElementsByClassName(self, clsName):
    """
    Description:
    ------------
    The getElementsByClassName() method returns a collection of all elements in the document with the specified class name, as a NodeList object.

    Related Pages:

			https://www.w3schools.com/jsref/met_document_getelementsbyclassname.asp

    Attributes:
    ----------
    :param clsName: Required. The class name of the elements you want to get.

    :return: A NodeList object, representing a collection of elements with the specified class name.
             The elements in the returned collection are sorted as they appear in the source code.
    """
    return "document.getElementsByClassName(%s)" % clsName

  def hasFocus(self):
    """
    Description:
    ------------
    The hasFocus() method returns a Boolean value indicating whether the document (or any element inside the document) has focus.

    Related Pages:

			https://www.w3schools.com/jsref/met_document_hasfocus.asp

    :return: A Boolean value, incidating whether the document or any element in the document has focus:
    """
    return "document.hasFocus()"

  def execCommand(self, command, showUI, value):
    """
    Description:
    ------------
    The execCommand() method executes the specified command for the selected part of an editable section.

    Related Pages:

			https://www.w3schools.com/jsref/met_document_execcommand.asp

    Attributes:
    ----------
    :param command:	 Specifies the name of the command to execute on the selected section
    :param showUI: A Boolean, specifies if the UI should be shown or not
    :param value: Some commands need a value to be completed

    :return: A Boolean, false if the command is not supported, otherwise true
    """
    return "document.execCommand('%s')" % command

  def createEvent(self, type):
    """
    Description:
    ------------
    The createEvent() method creates an event object.

    The event can be of any legal event type, and must be initialized before use.

    Related Pages:

			https://www.w3schools.com/jsref/event_createevent.asp

    Attributes:
    ----------
    :param type: Required. A String that specifies the type of the event.

    :return: An Event object
    """
    if type not in ['AnimationEvent', 'ClipboardEvent', 'DragEvent', 'FocusEvent', 'HashChangeEvent', 'InputEvent',
                    'MouseEvent', 'PageTransitionEvent', 'PopStateEvent', 'ProgressEvent', 'StorageEvent', 'TouchEvent',
                    'TransitionEvent', 'UiEvent', 'WheelEvent', 'KeyboardEvent']:
      raise Exception("Not recognized type %s" % type)

  def createAttribute(self, attributename):
    """
    Description:
    ------------
    The createAttribute() method creates an attribute with the specified name, and returns the attribute as an Attr object.

    Related Pages:

			https://www.w3schools.com/jsref/met_document_createattribute.asp

    Attributes:
    ----------
    :param attributename: Required. The name of the attribute you want to create

    :return: A Node object, representing the created attribute
    """
    return JsNodeAttributes.JsAttributes(attributename)

  def writeln(self, jsString):
    """
    Description:
    ------------
    The writeln() method is identical to the document.write() method, with the addition of writing a newline character after each statement.

    Related Pages:

			https://www.w3schools.com/jsref/met_doc_writeln.asp

    :param jsString: Optional. What to write to the output stream.
                     Multiple arguments can be listed and they will be appended to the document in order of occurrence

    :return: No return value
    """
    return "document.writeln(%s)" % jsString

  # -----------------------------------------------------------------
  #                         DATA TYPES
  #
  # def date(self, data=None, varName=None, setVariable=False):
  #   """
  #   Create or get a Javascript Date object.
  #
  #   Example
  #   rptObj.js.date(setVariable=False).getMonth()
  #
  #   Documentation
  #   https://www.w3schools.com/jsref/jsref_obj_date.asp
  #
  #   :param data: The object data
  #   :param varName: The array javascript reference
  #   :param setVariable: A flag to force the variable to be defined on the Javascript side
  #   :return: A Python JsDate object
  #   """
  #   if data is not None and varName is not None:
  #     setVariable = True
  #   return JsDate.JsDate(data, varName, setVariable)
  #
  # def array(self, data=None, varName=None, setVariable=False):
  #   """
  #   Create or get a Javascript Array object.
  #
  #   Example
  #   rptObj.js.array([], varName="myList")
  #
  #   Documentation
  #   https://www.w3schools.com/jsref/jsref_obj_array.asp
  #
  #   :param data: The object data
  #   :param varName: The array javascript reference
  #   :param setVariable: A flag to force the variable to be defined on the Javascript side
  #   :return: A Python JsArray object
  #   """
  #   if data is not None and varName is not None:
  #     setVariable = True
  #   return JsArray.JsArray(data, varName, setVariable)
  #
  # def object(self, data=None, varName=None, setVariable=False):
  #   """
  #
  #   :param varName:
  #   :param varType:
  #   :return:
  #   """
  #   if data is not None and varName is not None:
  #     setVariable = True
  #   return JsObject.JsObject(data, varName, setVariable)
  #
  # def number(self, data=None, varName=None, setVariable=False):
  #   """
  #   Create or get a Javascript Number object.
  #
  #   Example
  #   rptObj.js.number(0, varName="myNumber")
  #   rptObj.js.number(varName="myNumber") + 76
  #
  #   Documentation
  #   https://www.w3schools.com/jsref/jsref_obj_number.asp
  #
  #   :param data: The object data
  #   :param varName: The array javascript reference
  #   :param setVariable: A flag to force the variable to be defined on the Javascript side
  #   :return: A Python JsNumber object
  #   """
  #   if data is not None and varName is not None:
  #     setVariable = True
  #   return JsNumber.JsNumber(data, varName, setVariable)
  #
  # def string(self, data=None, varName=None, setVariable=False):
  #   """
  #
  #   :param data:
  #   :param varName:
  #   :param setVariable:
  #   :return: A Python JsString object
  #   """
  #   if data is not None and varName is not None:
  #     setVariable = True
  #   return JsString.JsString(JsUtils.jsConvertData(data, None), varName=varName, setVar=setVariable)

  def parseFloat(self, jsString):
    """
    Description:
    ------------
    The parseFloat() function parses a string and returns a floating point number.

    Related Pages:

			https://www.w3schools.com/jsref/jsref_parseint.asp

    Attributes:
    ----------
    :param jsString: Required. The string to be parsed

    :return: A Number. If the first character cannot be converted to a number, NaN is returned
    """
    return JsNumber.JsNumber("parseFloat(%s)" % jsString, isPyData=False)

  def parseInt(self, jsString):
    """
    Description:
    ------------
    The parseInt() function parses a string and returns an integer.

    Related Pages:

			https://www.w3schools.com/jsref/jsref_parseint.asp

    Attributes:
    ----------
    :param jsString: Required. The string to be parsed

    :return: A Number. If the first character cannot be converted to a number, NaN is returned
    """
    return JsNumber.JsNumber("parseInt(%s)" % jsString, isPyData=False)

  def parseDate(self, jsString):
    """
    Description:
    ------------
    The parse() method parses a date string and returns the number of milliseconds between the date string and midnight of January 1, 1970.

    Related Pages:

			https://www.w3schools.com/jsref/jsref_parse.asp

    Attributes:
    ----------
    :param jsString: Required. A string representing a date

    :return: A Number, representing the number of milliseconds between the specified date-time and midnight January 1, 1970

    """
    return JsNumber.JsNumber("Date.parse(%s)" % jsString, isPyData=False)

  def getVar(self, varName, varType="var"):
    """
    Description:
    ------------
    Get the Javascript Variable name

    Attributes:
    ----------
    :param varName: The Variable name
    :param varType: The scope of the variable

    :return: Return the piece of script to be added to the Javascript
    """
    if varType == 'var':
      return "window['%s']" % varName

    return varName

  def addToDataDict(self, key, jsVal, jsDataKey=None, isPyData=True, jsFnc=None, jsData="data"):
    """

    Attributes:
    ----------
    :param key:
    :param jsVal:
    :param varType:
    :param jsDataKey:
    :param isPyData:
    :param jsFnc:
    :param jsData:
    """
    jsVal = JsUtils.jsConvert(jsVal, jsDataKey, isPyData, jsFnc)
    return "%s['%s'] = %s" % (jsData, key, jsVal)

  def getItem(self, htmlCode):
    """

    :param htmlCode:

    :return:
    """
    if not htmlCode in self._context.get('htmlCodes', []):
      raise Exception("HTML Code - %s - not defined in the current report" % htmlCode)

    return "$('#%s')" % htmlCode

  def typeof(self, jsData):
    """

    :param jsData:

    :return:
    """
    return JsFncs.JsTypeOf(JsUtils.jsConvertData(jsData, None))

  def info(self, jsData, cssStyle=None, icon="fas fa-spinner fa-spin", seconds=10000):
    """
    Description:
    ------------
    Display a message

    Related Pages:

			https://fontawesome.com/how-to-use/on-the-web/styling/animating-icons

    Attributes:
    ----------
    :param jsData:
    :param cssStyle:
    :param icon:
    :param seconds:
    """
    if cssStyle is None:
      cssStyle = {"position": "fixed", "bottom": "5px", "right": "10px", "padding": '2px 7px', "border": "1px solid black"}
    if icon is not None:
      self._src.jsImports.add('font-awesome')
      self._src.cssImport.add('font-awesome')
      return [
        self.createElement("i", varName="popup_icon").setAttribute("aria-hidden", True).css({"display": "inline-block",
                           "width": "auto", "height": "auto", "margin-right": '5px'}).className(icon),
        self.createElement("div", varName="popup_info").appendChild(self.objects.dom("popup_icon")).css(cssStyle).text(jsData),
        self.body.appendChild(self.objects.dom("popup_info")),
        self.window.setTimeout(self.objects.dom("popup_info").remove(), milliseconds=seconds)]

    return [
      self.createElement("div", varName="popup_info").css(cssStyle).text(jsData),
      self.body.appendChild(self.objects.dom("popup_info")),
      self.window.setTimeout(self.objects.dom("popup_info").remove(), milliseconds=seconds)]




from epyk.interfaces.geo import CompGeoPlotly
from epyk.interfaces.geo import CompGeoDc
from epyk.interfaces.geo import CompGeoChartJs


class Geo(object):
  def __init__(self, context):
    self.context = context

  @property
  def plotly_map(self):
    """
    Interface for the Plotly library

    :return: CompGeoPlotly.Plotly
    """
    return CompGeoPlotly.Plotly(self)

  @property
  def dc_choropleth(self):
    """

    :return: CompGeoDc.Dc
    """
    return CompGeoDc.Dc(self)

  @property
  def chartjs_choropleth(self):
    """

    :return: CompGeoChartJs.ChartJs
    """
    return CompGeoChartJs.ChartJs(self)

  @property
  def plotly_choropleth(self):
    """

    :return: CompGeoPlotly.PlotlyChoropleth
    """
    return CompGeoPlotly.PlotlyChoropleth(self)

  @property
  def plotly_scatter(self):
    """

    https://plot.ly/javascript/scatter-plots-on-maps/

    :return: CompGeoPlotly.PlotlyScatter
    """
    return CompGeoPlotly.PlotlyScatter(self)

  @property
  def plotly_bubble(self):
    """

    https://plot.ly/javascript/bubble-maps/

    :return: CompGeoPlotly.PlotlyBubble
    """
    return CompGeoPlotly.PlotlyBubble(self)




from epyk.core import html
from epyk.core.css import Defaults_css


class Menus(object):
  def __init__(self, context):
    self.context = context

  def top(self, data=None, color=None, width=(100, "%"), height=(None, 'px'),
           htmlCode=None, helper=None, options=None, profile=None):
    """
    Usage::

      l = rptObj.ui.lists.list(["A", "B"])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`
      - :class:`epyk.core.html.HtmlContainer.Col`
      - :class:`epyk.core.html.HtmlContainer.Grid`
      - :class:`epyk.core.html.HtmlText.Title`
      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

        https://www.w3schools.com/bootstrap/bootstrap_list_groups.asp
    http://astronautweb.co/snippet/font-awesome/
    """
    menu_li, menu_title, menu_items = [], [], []
    for k in data:
      menu_li.append(k["value"])
      menu_title.append(k.get("title"))
      menu_items.append(k.get("children", []))
    html_list = html.HtmlList.List(self.context.rptObj, menu_li, color, width, height, htmlCode,
                                   helper, options or {}, profile)
    html_list.css({"list-style": 'none'})
    html_div = self.context.rptObj.ui.div(
      self.context.rptObj.ui.grid([
        self.context.rptObj.ui.col([
          self.context.rptObj.ui.title("test", level=1),
          *menu_items[0]
        ]).css({"color": "white", "padding": "0 5px"})
      ])
    )

    col = self.context.rptObj.ui.col([html_list, html_div])
    col.css({"background-color": "#333", "position": "fixed", "margin": 0, "top": 0, "left": 0, "color": 'white'})
    self.context.register(col)
    return col

  def bottom(self, data=None, color=None, width=(100, "%"), height=(None, 'px'),
           htmlCode=None, helper=None, options=None, profile=None):
    """
    Usage::

      l = rptObj.ui.lists.list(["A", "B"])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`
      - :class:`epyk.core.html.HtmlContainer.Col`
      - :class:`epyk.core.html.HtmlText.Title`
      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

        https://www.w3schools.com/bootstrap/bootstrap_list_groups.asp
    http://astronautweb.co/snippet/font-awesome/
    """
    menu_li, menu_title, menu_items = [], [], []
    for k in data:
      menu_li.append(k["value"])
      title_text = k.get("title")
      if title_text is not None:
        title_text = self.context.rptObj.ui.title(title_text, level=4)
        title_text.inReport = False
      menu_title.append(title_text)
      menu_items.append(k.get("children", []))
    html_list = self.context.rptObj.ui.list(menu_li, color, width, height, htmlCode, helper, options or {}, profile)
    html_list.css({"list-style": 'none'})
    html_div = self.context.rptObj.ui.div(
      self.context.rptObj.ui.grid([
        self.context.rptObj.ui.col([
          menu_title[0], *menu_items[0]]).css({"color": "white", "padding": "0 5px"})
      ])
    )
    html_div.style.display = None

    html_list.click_items([
      #self.context.rptObj.js.console.log(html_list),
      html_div.dom.toggle()
    ])

    col = self.context.rptObj.ui.col([html_div, html_list])
    col.css({"background-color": "#333",
             "position": "fixed", "bottom": 0, "left": 0,
             "margin": 0, "color": 'white'})
    return col

  def menu(self, data=None, position="bottom", color=None, width=(100, "%"), height=(None, 'px'),
           htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      l = rptObj.ui.lists.list(["A", "B"])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`
      - :class:`epyk.core.html.HtmlContainer.Col`
      - :class:`epyk.core.html.HtmlContainer.Grid`
      - :class:`epyk.core.html.HtmlText.Title`
      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

        https://www.w3schools.com/bootstrap/bootstrap_list_groups.asp
    http://astronautweb.co/snippet/font-awesome/
    """
    menu_li, menu_title, menu_items, menu_divs = [], [], [], []
    for k in data:
      menu_li.append(k["value"])
      title_text = k.get("title")
      if title_text is not None:
        title_text = self.context.rptObj.ui.title(title_text, level=4)
        title_text.inReport = False
      menu_title.append(title_text)
      menu_items.append(k.get("children", []))
    html_list = self.context.rptObj.ui.list(menu_li, color, width, height, htmlCode, helper, options or {}, profile)
    html_list.css({"list-style": 'none'})
    for i, m in enumerate(menu_title):
      if isinstance(menu_items[i][0], list):
        grid = self.context.rptObj.ui.div([])
        for item in menu_items[i]:
          grid + self.context.rptObj.ui.col([m, *item], width=(None, "px")).css({"color": "white", "padding": "0 5px",
                        "display": 'inline-block', "vertical-align": 'top', "margin": '2px 0'})
        html_div = self.context.rptObj.ui.div(grid).css({"vertical-align": 'None'})
        html_div.attr["name"] = "divs_%s" % (html_list.htmlId)
        html_div.style.display = None
      else:
        html_div = self.context.rptObj.ui.div(
          self.context.rptObj.ui.grid([
            self.context.rptObj.ui.col([m, *menu_items[i]]).css({"color": "white", "padding": "0 5px"})]))
        html_div.attr["name"] = "divs_%s" % (html_list.htmlId)
        html_div.style.display = None
      menu_divs.append(html_div)
    html_list.click_items(
      [self.context.rptObj.js.getElementById(l.htmlId).setAttribute("data-select", "false") for l in html_list]
      + [self.context.rptObj.js.objects.dom("this").setAttribute("data-select", "true")])
    col = self.context.rptObj.ui.col([html_list, *menu_divs])
    col.css({"background-color": "#333", "margin": 0, "color": 'white'})
    return col

  def icons(self):
    pass

  def buttons(self, data=None, color=None, width=(100, "%"), height=(None, 'px'),
              htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      bs = rptObj.ui.buttons.buttons(["Button", "Button 2", "Button 3"])
      bs[2].click([
      rptObj.js.alert(bs[2].dom.content)
      ])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlButton.Buttons`

    Attributes:
    ----------
    :param data:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param profile:
    """
    dfl_button_css = {"button_css": {"border-radius": 0, "border": "0px solid black"}}
    options = options or {}
    dfl_button_css.update(options)
    html_obj = html.HtmlButton.Buttons(self.context.rptObj, data or [], color, width, height, htmlCode, helper,
                                       dfl_button_css, profile)
    html_obj.css({"border": "1px solid %s" % html_obj._report.theme.greys[4], "padding": "2px"})
    self.context.register(html_obj)
    return html_obj

  def right(self):
    pass

  def left(self):
    pass

  def divisor(self, data, divider=None, width=(100, '%'), height=(None, 'px'), options=None, profile=False):
    """
    Description:
    ------------

    Usage::

      record = []
      rptObj.ui.menus.divisor(record)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`
      - :class:`epyk.core.html.HtmlText.Text`

    Attributes:
    ----------
    :param divider:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    if divider is None:
      divider = self.context.rptObj.symbols.shapes.BLACK_RIGHT_POINTING_TRIANGLE
    div = self.context.rptObj.ui.div(width=width, height=height, options=options, profile=profile)
    div.texts = []
    for rec in data[:-1]:
      div.texts.append(self.context.rptObj.ui.text(rec).css({"display": 'inline-block'}))
      div += div.texts[-1]
      div += self.context.rptObj.ui.text(divider).css({"display": 'inline-block', 'margin': '0 5px', 'font-size': Defaults_css.font(-2)})
    div +=self.context.rptObj.ui.text(data[-1]).css({"display": 'inline-block'})
    return div

  def button(self, value, object, symbol=None, width=("auto", ''), height=(None, 'px'), options=None, profile=False):
    """
    Description:
    ------------

    Usage::

      mb = rptObj.ui.menus.button("Value", rptObj.ui.button("sub button"))
      mb.item.click([rptObj.js.alert(mb.item.dom.content)])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`
      - :class:`epyk.core.html.HtmlButton.Button`

    Attributes:
    ----------
    :param value:
    :param object:
    :param symbol:
    :param width:
    :param height:
    :param options:
    :param profile:
    """

    div = self.context.rptObj.ui.div(width=width, height=height, options=options, profile=profile)
    div.item = object
    content = self.context.rptObj.ui.div(object, width=("auto", ''))
    content.style.css.display = None
    content.style.css.bottom = 0
    content.style.css.padding = 5
    content.style.css.background = self.context.rptObj.theme.greys[0]
    content.style.css.z_index = 5
    content.style.css.position = 'absolute'
    if symbol is None:
      symbol = self.context.rptObj.symbols.shapes.BLACK_DOWN_POINTING_SMALL_TRIANGLE
    but = self.context.rptObj.ui.button("%s %s" % (value, symbol),
                                        width=width, profile=profile)
    div += but
    div += content
    div.on("mouseover", [content.dom.css({"display": 'block'})])
    div.on("mouseout", [content.dom.css({"display": 'none'})])
    return div

  def toolbar(self, data, width=("auto", ''), height=(None, 'px'), options=None, profile=False):
    """
    Description:
    ------------

    Usage::

      tb = rptObj.ui.menus.toolbar(["fas fa-paint-brush", "fas fa-code"])
      tb[1].link.val = 4589
      tb[1].tooltip("This is a tooltip")
      tb[0].style.css.color = 'red'

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`
      - :class:`epyk.core.html.HtmlImage.Badge`

    Attributes:
    ----------
    :param data:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    div = self.context.rptObj.ui.div(width=width, height=height, options=options, profile=profile)
    for d in data:
      div += self.context.rptObj.ui.images.badge(icon=d, text="", options={"badge_position": 'right'})
      div[-1].style.css.cursor = 'pointer'
    return div

  def selections(self, data, width=(100, '%'), height=(20, 'px'), htmlCode=None, attrs=None,
                  helper=None, options=None, profile=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlEvent.Menu`

    :param data:
    :param width:
    :param height:
    :param htmlCode:
    :param attrs:
    :param helper:
    :param options:
    :param profile:
    :return:
    """
    html_pr = html.HtmlEvent.Menu(self.context.rptObj, data, width, height,  attrs or {}, helper,
                                         options or {}, htmlCode, profile)
    self.context.register(html_pr)
    return html_pr



"""
Related Pages:

		https://www.w3schools.com/charsets/ref_utf_arrows.asp
"""

LEFTWARDS_ARROW = '&#8592'
UPWARDS_ARROW = '&#8593'
RIGHTWARDS_ARROW = '&#8594'
DOWNWARDS_ARROW = '&#8595'
LEFT_RIGHT_ARROW = '&#8596'
UP_DOWN_ARROW = '&#8597'
NORTH_WEST_ARROW = '&#8598'
NORTH_EAST_ARROW = '&#8599'
SOUTH_EAST_ARROW = '&#8600'
SOUTH_WEST_ARROW = '&#8601'
LEFTWARDS_ARROW_WITH_STROKE = '&#8602'
RIGHTWARDS_ARROW_WITH_STROKE = '&#8603'
LEFTWARDS_WAVE_ARROW = '&#8604'
RIGHTWARDS_WAVE_ARROW = '&#8605'
LEFTWARDS_TWO_HEADED_ARROW = '&#8606'
UPWARDS_TWO_HEADED_ARROW = '&#8607'
RIGHTWARDS_TWO_HEADED_ARROW = '&#8608'
DOWNWARDS_TWO_HEADED_ARROW = '&#8609'
LEFTWARDS_ARROW_WITH_TAIL = '&#8610'
RIGHTWARDS_ARROW_WITH_TAIL = '&#8611'
LEFTWARDS_ARROW_FROM_BAR = '&#8612'
UPWARDS_ARROW_FROM_BAR = '&#8613'
RIGHTWARDS_ARROW_FROM_BAR = '&#8614'
DOWNWARDS_ARROW_FROM_BAR = '&#8615'
UP_DOWN_ARROW_WITH_BASE = '&#8616'
LEFTWARDS_ARROW_WITH_HOOK = '&#8617'
RIGHTWARDS_ARROW_WITH_HOOK = '&#8618'
LEFTWARDS_ARROW_WITH_LOOP = '&#8619'
RIGHTWARDS_ARROW_WITH_LOOP = '&#8620'
LEFT_RIGHT_WAVE_ARROW = '&#8621'
LEFT_RIGHT_ARROW_WITH_STROKE = '&#8622'
DOWNWARDS_ZIGZAG_ARROW = '&#8623'
UPWARDS_ARROW_WITH_TIP_LEFTWARDS = '&#8624'
UPWARDS_ARROW_WITH_TIP_RIGHTWARDS = '&#8625'
DOWNWARDS_ARROW_WITH_TIP_LEFTWARDS = '&#8626'
DOWNWARDS_ARROW_WITH_TIP_RIGHTWARDS = '&#8627'
RIGHTWARDS_ARROW_WITH_CORNER_DOWNWARDS = '&#8628'
DOWNWARDS_ARROW_WITH_CORNER_LEFTWARDS = '&#8629'
ANTICLOCKWISE_TOP_SEMICIRCLE_ARROW = '&#8630'
CLOCKWISE_TOP_SEMICIRCLE_ARROW = '&#8631'
NORTH_WEST_ARROW_TO_LONG_BAR = '&#8632'
LEFTWARDS_ARROW_TO_BAR_OVER_RIGHTWARDS_ARROW_TO_BAR = '&#8633'
ANTICLOCKWISE_OPEN_CIRCLE_ARROW = '&#8634'
CLOCKWISE_OPEN_CIRCLE_ARROW = '&#8635'
LEFTWARDS_HARPOON_WITH_BARB_UPWARDS = '&#8636'
LEFTWARDS_HARPOON_WITH_BARB_DOWNWARDS = '&#8637'
UPWARDS_HARPOON_WITH_BARB_RIGHTWARDS = '&#8638'
UPWARDS_HARPOON_WITH_BARB_LEFTWARDS = '&#8639'
RIGHTWARDS_HARPOON_WITH_BARB_UPWARDS = '&#8640'
RIGHTWARDS_HARPOON_WITH_BARB_DOWNWARDS = '&#8641'
DOWNWARDS_HARPOON_WITH_BARB_RIGHTWARDS = '&#8642'
DOWNWARDS_HARPOON_WITH_BARB_LEFTWARDS = '&#8643'
RIGHTWARDS_ARROW_OVER_LEFTWARDS_ARROW = '&#8644'
UPWARDS_ARROW_LEFTWARDS_OF_DOWNWARDS_ARROW = '&#8645'
LEFTWARDS_ARROW_OVER_RIGHTWARDS_ARROW = '&#8646'
LEFTWARDS_PAIRED_ARROWS = '&#8647'
UPWARDS_PAIRED_ARROWS = '&#8648'
RIGHTWARDS_PAIRED_ARROWS = '&#8649'
DOWNWARDS_PAIRED_ARROWS = '&#8650'
LEFTWARDS_HARPOON_OVER_RIGHTWARDS_HARPOON = '&#8651'
RIGHTWARDS_HARPOON_OVER_LEFTWARDS_HARPOON = '&#8652'
LEFTWARDS_DOUBLE_ARROW_WITH_STROKE = '&#8653'
LEFT_RIGHT_DOUBLE_ARROW_WITH_STROKE = '&#8654'
RIGHTWARDS_DOUBLE_ARROW_WITH_STROKE = '&#8655'
LEFTWARDS_DOUBLE_ARROW = '&#8656'
UPWARDS_DOUBLE_ARROW = '&#8657'
RIGHTWARDS_DOUBLE_ARROW = '&#8658'
DOWNWARDS_DOUBLE_ARROW = '&#8659'
LEFT_RIGHT_DOUBLE_ARROW = '&#8660'
UP_DOWN_DOUBLE_ARROW = '&#8661'
NORTH_WEST_DOUBLE_ARROW = '&#8662'
NORTH_EAST_DOUBLE_ARROW = '&#8663'
SOUTH_EAST_DOUBLE_ARROW = '&#8664'
SOUTH_WEST_DOUBLE_ARROW = '&#8665'
LEFTWARDS_TRIPLE_ARROW = '&#8666'
RIGHTWARDS_TRIPLE_ARROW = '&#8667'
LEFTWARDS_SQUIGGLE_ARROW = '&#8668'
RIGHTWARDS_SQUIGGLE_ARROW = '&#8669'
UPWARDS_ARROW_WITH_DOUBLE_STROKE = '&#8670'
DOWNWARDS_ARROW_WITH_DOUBLE_STROKE = '&#8671'
LEFTWARDS_DASHED_ARROW = '&#8672'
UPWARDS_DASHED_ARROW = '&#8673'
RIGHTWARDS_DASHED_ARROW = '&#8674'
DOWNWARDS_DASHED_ARROW = '&#8675'
LEFTWARDS_ARROW_TO_BAR = '&#8676'
RIGHTWARDS_ARROW_TO_BAR = '&#8677'
LEFTWARDS_WHITE_ARROW = '&#8678'
UPWARDS_WHITE_ARROW = '&#8679'
RIGHTWARDS_WHITE_ARROW = '&#8680'
DOWNWARDS_WHITE_ARROW = '&#8681'
UPWARDS_WHITE_ARROW_FROM_BAR = '&#8682'
UPWARDS_WHITE_ARROW_ON_PEDESTAL = '&#8683'
UPWARDS_WHITE_ARROW_ON_PEDESTAL_WITH_HORIZONTAL_BAR = '&#8684'
UPWARDS_WHITE_ARROW_ON_PEDESTAL_WITH_VERTICAL_BAR = '&#8685'
UPWARDS_WHITE_DOUBLE_ARROW = '&#8686'
UPWARDS_WHITE_DOUBLE_ARROW_ON_PEDESTAL = '&#8687'
RIGHTWARDS_WHITE_ARROW_FROM_WALL = '&#8688'
NORTH_WEST_ARROW_TO_CORNER = '&#8689'
SOUTH_EAST_ARROW_TO_CORNER = '&#8690'
UP_DOWN_WHITE_ARROW = '&#8691'
RIGHT_ARROW_WITH_SMALL_CIRCLE = '&#8692'
DOWNWARDS_ARROW_LEFTWARDS_OF_UPWARDS_ARROW = '&#8693'
THREE_RIGHTWARDS_ARROWS = '&#8694'
LEFTWARDS_ARROW_WITH_VERTICAL_STROKE = '&#8695'
RIGHTWARDS_ARROW_WITH_VERTICAL_STROKE = '&#8696'
LEFT_RIGHT_ARROW_WITH_VERTICAL_STROKE = '&#8697'
LEFTWARDS_ARROW_WITH_DOUBLE_VERTICAL_STROKE = '&#8698'
RIGHTWARDS_ARROW_WITH_DOUBLE_VERTICAL_STROKE = '&#8699'
LEFT_RIGHT_ARROW_WITH_DOUBLE_VERTICAL_STROKE = '&#8700'
LEFTWARDS_OPEN_HEADED_ARROW = '&#8701'
RIGHTWARDS_OPEN_HEADED_ARROW = '&#8702'
LEFT_RIGHT_OPEN_HEADED_ARROW = '&#8703'



"""
Core module in charge of linking the Python report request to the corresponding external pacakges required.

This package will resolve the external Javascript and CSS dependencies.

It can also help or retrieve external Python packages from the official Python Page Index online repository
"""

import re
import os
import sys
import json
import importlib
import collections

# To fully disable the automatic pip install request when a package is missing
AUTOLOAD = False
PROXY = ''
PCK_REPO = ''
STATIC_PATH = "/static"


def requires(name, reason='Missing Package', install=None, package=None, raise_except=False, source_script=None, pip_attrs=None):
  """
  Description:
  ------------
  System module

  Import the necessary external packages and provide explicit message to find a way to solve this error message.
  This method should also explain why this module is required to make sure this is really expected to get an error.

  Attributes:
  ----------
  :param name:
  :param reason:
  :param install:
  :param package:
  :param raise_except:
  :param source_script:
  :param pip_attrs: Optional. The pip attributes  https://packaging.python.org/tutorials/installing-packages/

  return: The python module
  """
  if install is None:
    install = name
  try:
    mod = importlib.import_module(name)
    if package is not None:
      return getattr(mod, package)

    return mod

  except Exception as err:
    if pip_attrs is None:
      pip_attrs = []
    if PROXY:
      pip_attrs.extend(['--proxy', PROXY])
    if PCK_REPO:
      pip_attrs.extend(['--no-index', '--find-links', PCK_REPO])

    if str(err).startswith("Missing required dependencies") and AUTOLOAD:
      print("Error with %s in script %s, autoload set to %s" % (name, source_script, AUTOLOAD))
      deps = json.loads(str(err).replace("Missing required dependencies ", "").replace("'", '"'))
      import subprocess
      for d in deps:
        exe_out = subprocess.call([sys.executable, '-m', "pip", 'install'] + pip_attrs + [d])
        print(exe_out)
      return requires(name, reason, install, package=package, raise_except=raise_except)

    if AUTOLOAD:
      if isinstance(AUTOLOAD, dict) and not AUTOLOAD.get(install, False):
        # Module not set in the configuration to be automatically loaded
        raise Exception(err)

      print("Error with %s in script %s, autoload set to %s" % (name, source_script, AUTOLOAD))
      import subprocess
      subprocess.call([sys.executable, '-m', "pip", 'install'] + pip_attrs + [install])

      return requires(name, reason, install, package=package, raise_except=raise_except)

    if raise_except:
      print("Error with %s in script %s, autoload set to %s" % (name, source_script, AUTOLOAD))
      print("*** Module %s required ***" % name)
      print(reason)
      if install:
        print("Command to fix this error:")
        print(">>> pip install %s" % install)
      print('')
      raise Exception(err)


def load_package(package_name, pip_attrs=None, action='install'):
  """
  Description:
  ------------
  Force the package to be installed manually to the currently python distribution.
  This will run a pip command to the running python set up

  Example
  load_package("pandas")

  Related Pages:

			https://pypi.org/

  Attributes:
  ----------
  :param package_name: The external package reference (e.g. pandas)
  :param pip_attrs: Optional. The pip attributes  https://packaging.python.org/tutorials/installing-packages/
  :param action: Optional. The pip command (default install)
  """
  import subprocess

  if pip_attrs is None:
    pip_attrs = ['--ignore-installed', '--upgrade'] if action == 'install' else []
  if PROXY is not None:
    subprocess.call([sys.executable, '-m', "pip", action, '--proxy="%s"' % PROXY] + pip_attrs + [package_name])
  else:
    subprocess.call([sys.executable, '-m', "pip", action] + pip_attrs + [package_name])


def installed_packages():
  """
  Description:
  ------------
  Returns the list of packages installed on the running Python distribution

  This will require an internet connection as it will run the pip command behind the scene.
  It will return in the console a table with the status of the obsolescence of all the python packages

  Example
  installed_packages()
  """
  import subprocess
  subprocess.call(["pip", 'list', '-o'])


CDNJS_REPO = 'https://cdnjs.cloudflare.com/ajax/libs'


JS_IMPORTS = {
  # numbers formatting
  'accounting': {
    'register': {'alias': 'accounting', 'module': 'accounting.min', 'name': 'accounting'},
    'modules': [
      # Better to use the bundle version to avoid the import issue with popper.js
      {'script': 'accounting.min.js', 'version': '0.4.1', 'path': 'accounting.js/%(version)s/', 'cdnjs': CDNJS_REPO},
    ],
    'website': 'https://openexchangerates.github.io/accounting.js/'},

  # Plolyfill
  'promise-polyfill': {
    'modules': [
      # Better to use the bundle version to avoid the import issue with popper.js
      {'script': 'polyfill.min.js', 'version': '4.2.1', 'path': 'promise-polyfill@8/dist/', 'cdnjs': 'https://cdn.jsdelivr.net/npm'},
    ],
    'website': 'https://github.com/taylorhakes/promise-polyfill'},

  # Common module for browser versions compatibilities
  'babel-polyfill': {
    'website': 'https://babeljs.io/',
    'register': {'alias': 'babel', 'module': 'polyfill', 'name': 'babel'},
    'modules': [
      {'script': 'polyfill.js', 'version': '7.4.4', 'path': 'babel-polyfill/%(version)s/', 'cdnjs': CDNJS_REPO},
    ]
  },

  # module are written from the first one to load to the last one
  'bootstrap': {
    'register': {'alias': 'bootstrap', 'module': 'bootstrap.min', 'name': 'bootstrap'},
    'req': [{'alias': 'jquery'}, {'alias': 'popper'}],
    'modules': [
      # Better to use the bundle version to avoid the import issue with popper.js
      {'script': 'bootstrap.min.js', 'version': '4.4.1', 'path': 'bootstrap/%(version)s/js/', 'cdnjs': 'https://stackpath.bootstrapcdn.com'},
    ],
    'website': 'https://getbootstrap.com/'},

  'moment': {
    'dsc': 'Module used by Tabulator for datetime objects',
    #'register': {'alias': 'moment', 'module': 'moment.min', 'npm': 'moment'},
    'modules': [
      {'script': 'moment.min.js', 'version': '2.24.0', 'path': 'moment.js/%(version)s/', 'cdnjs': CDNJS_REPO},
    ],
    'website': 'https://momentjs.com/',
  },

  # AG Grid tables
  'ag-grid': {
    'website': 'https://www.ag-grid.com/javascript-grid/',
    'modules': [
      {'script': 'ag-grid-community.min.js', 'version': '23.0.2', 'path': 'ag-grid/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # module for tabulator
  'tabulator': {
    'req': [{'alias': 'promise-polyfill'}, {'alias': 'moment'}],
    'register': {'alias': 'Tabulator', 'module': 'tabulator.min', 'npm': 'tabulator'},
    'modules': [
      # core only needed for Jupyter for some reasons
      {'script': 'tabulator_core.min.js', 'version': '4.4.3', 'path': 'tabulator/%(version)s/js/', 'cdnjs': CDNJS_REPO}, # 'reqAlias': 'tabulator_core',
      {'script': 'tabulator.min.js', 'version': '4.4.3', 'path': 'tabulator/%(version)s/js/', 'cdnjs': CDNJS_REPO}
    ],
    'website': 'http://tabulator.info/'
  },

  'tabulator-inputs': {
    'req': [{'alias': 'tabulator'}],
    'modules': [
      {'script': 'formatters-inputs.js', 'version': 'NA', 'path': 'formatters/', 'cdnjs': 'https://raw.githubusercontent.com/epykure/tabulator-extensions/master'},
    ],
    'website': 'http://tabulator.info/'
  },

  'tabulator-mutators-inputs': {
    'req': [{'alias': 'tabulator'}],
    'modules': [
      {'script': 'mutators-inputs.js', 'version': 'NA', 'path': 'mutators/', 'cdnjs': 'https://raw.githubusercontent.com/epykure/tabulator-extensions/master'},
    ],
    'website': 'http://tabulator.info/'
  },

  'editors-inputs': {
    'req': [{'alias': 'tabulator'}],
    'modules': [
      {'script': 'editors-inputs.js', 'version': 'NA', 'path': 'editors/', 'cdnjs': 'https://raw.githubusercontent.com/epykure/tabulator-extensions/master'}],
    'website': 'http://tabulator.info/'
  },

  'editors-dates': {
    'req': [{'alias': 'tabulator'}],
    'modules': [
      {'script': 'editors-dates.js', 'version': 'NA', 'path': 'editors/', 'cdnjs': 'https://raw.githubusercontent.com/epykure/tabulator-extensions/master'},
    ],
    'website': 'http://tabulator.info/'
  },

  'editors-selects': {
    'req': [{'alias': 'tabulator'}],
    'modules': [
      {'script': 'editors-selects.js', 'version': 'NA', 'path': 'editors/', 'cdnjs': 'https://raw.githubusercontent.com/epykure/tabulator-extensions/master'},
    ],
    'website': 'http://tabulator.info/'
  },

  'tabulator-icons': {
    'req': [{'alias': 'tabulator'}],
    'modules': [
      # core only needed for Jupyter for some reasons
      {'script': 'formatters-icons.js', 'version': 'NA', 'path': 'formatters/', 'cdnjs': 'https://raw.githubusercontent.com/epykure/tabulator-extensions/master'},
    ],
    'website': 'http://tabulator.info/'
  },

  'tabulator-editors': {
      'req': [{'alias': 'tabulator'}],
      'modules': [
        # core only needed for Jupyter for some reasons
        {'script': 'editors-bespoke.js', 'version': 'NA', 'path': 'editors/', 'cdnjs': 'https://raw.githubusercontent.com/epykure/tabulator-extensions/master'},
      ],
      'website': 'http://tabulator.info/'
    },

  'tabulator-numbers': {
      'req': [
        {'alias': 'tabulator'},
        {'alias': 'accounting'},
        {'alias': 'd3-packages'},
      ],
      'modules': [
        # core only needed for Jupyter for some reasons
        {'script': 'formatters-numbers.js', 'version': 'NA', 'path': 'formatters/', 'cdnjs': 'https://raw.githubusercontent.com/epykure/tabulator-extensions/master'},
      ],
      'website': 'http://tabulator.info/'
    },

  # module for the awesome icons
  'font-awesome': {
    'register': {'alias': 'fontawesome', 'module': 'fontawesome', 'npm': '@fortawesome/fontawesome-free'},
    'package': {'zip': 'https://use.fontawesome.com/releases/v%(version)s/fontawesome-free-%(version)s-web.zip',
                'root': 'fontawesome-free-%(version)s-web', 'folder': 'releases', 'path': 'v%(version)s'},
    'modules': [{'script': 'fontawesome.js', 'version': '5.11.2', 'path': 'releases/v%(version)s/js/', 'cdnjs': 'https://use.fontawesome.com'}],
    'website': 'https://fontawesome.com/'},

  # Javascript packages to handle DataTables
  'datatables': {
    'req': [{'alias': 'jquery'}],
    'register': {'alias': 'datatables', 'module': 'jquery.dataTables.min'},
    'modules': [
      {'reqAlias': 'datatables', 'script': 'jquery.dataTables.min.js', 'version': '1.10.19', 'path': 'datatables/%(version)s/js/', 'cdnjs': CDNJS_REPO},
  ]},

  # Datatable Buttons
  'datatables-buttons': {
    'website': 'https://datatables.net/extensions/buttons/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'dataTables.buttons.min.js', 'version': '1.6.1', 'path': 'buttons/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable Select
  'datatables-select': {
    'website': 'https://datatables.net/extensions/select/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'dataTables.select.min.js', 'version': '1.3.1', 'path': 'select/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable Scroller
  'datatables-scroller': {
    'website': 'https://datatables.net/extensions/scroller/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'dataTables.scroller.min.js', 'version': '2.0.1', 'path': 'scroller/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable SearchPanes
  'datatables-searchPanes': {
    'website': 'https://datatables.net/extensions/searchPanes/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'dataTables.searchPanes.min.js', 'version': '1.0.1', 'path': 'searchpanes/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable responsive
  'datatables-responsive': {
    'website': 'https://datatables.net/extensions/responsive/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'dataTables.responsive.min.js', 'version': '2.2.3', 'path': 'responsive/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable KeyTable
  'datatables-keytable': {
    'website': 'https://datatables.net/extensions/keytable/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'dataTables.keyTable.min.js', 'version': '2.5.1', 'path': 'keytable/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable autoFill
  'datatables-autoFill': {
    'website': 'https://datatables.net/extensions/autofill/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'dataTables.autoFill.min.js', 'version': '2.1.0', 'path': 'autofill/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable group rows
  'datatables-rows-group': {
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'dataTables.rowsGroup.js', 'version': '1.0.0', 'path': 'datatables-rowsgroup/v%(version)s/', 'cdnjs': 'https://cdn.rawgit.com/ashl1'}],
       'website': 'https://datatables.net/forums/discussion/29319/new-rowsgroup-plugin-merge-cells-vertically-rowspan'},

  # Datatable group row
  'datatables-row-group': {
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'dataTables.rowGroup.min.js', 'version': '1.1.1', 'path': 'rowgroup/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'}],
    'website': 'https://datatables.net/extensions/rowgroup/'},

  # Datatable fixed column
  'datatables-fixed-columns': {
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'dataTables.fixedColumns.min.js', 'version': '3.2.2', 'path': 'fixedcolumns/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'}],
    'website': 'https://datatables.net/extensions/fixedcolumns/'},

  # Datatable Fixed header
  'datatables-fixed-header': {
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'dataTables.fixedHeader.min.js', 'version': '3.1.3', 'path': 'fixedheader/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'}],
    'website': 'https://datatables.net/extensions/fixedheader/'},

  # Datatable data export
  'datatables-export': {
    'req': [
      {'alias': 'datatables'},
      {'alias': 'jszip'},
      {'alias': 'pdfmake'}],
    'website': 'https://datatables.net/extensions/buttons/',
    'modules': [
      {'script': 'buttons.colVis.min.js', 'version': '1.5.2', 'path': 'buttons/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'},
      {'script': 'buttons.bootstrap4.min.js', 'version': '1.5.2', 'path': 'buttons/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'},
      {'script': 'buttons.foundation.min.js', 'version': '1.5.2', 'path': 'buttons/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'},
      {'script': 'buttons.html5.min.js', 'version': '1.5.2', 'path': 'buttons/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'},
      {'script': 'buttons.jqueryui.min.js', 'version': '1.5.2', 'path': 'buttons/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'},
      {'script': 'buttons.print.min.js', 'version': '1.5.2', 'path': 'buttons/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'},
      {'script': 'buttons.semanticui.min.js', 'version': '1.5.2', 'path': 'buttons/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable column reordering modules
  'datatables-col-order': {
    'req': [{'alias': 'datatables'}],
    'register': {'alias': 'datatableColOrd', 'module': 'dataTables.colReorder.min'},
    'website': 'https://datatables.net/extensions/colreorder/',
    'modules': [
      {'reqAlias': 'datatableColOrd', 'script': 'dataTables.colReorder.min.js', 'version': '1.5.1', 'path': 'colreorder/%(version)s/js/', 'cdnjs': 'https://cdn.datatables.net'}]},

  #
  'jszip': {
    'website': 'https://datatables.net/extensions/buttons/',
    'register': {'alias': 'jszip', 'module': 'jszip.min'},
    'modules': [
      {'reqAlias': 'jszip', 'script': 'jszip.min.js', 'version': '3.1.5', 'path': 'jszip/%(version)s/', 'cdnjs': CDNJS_REPO},
    ]},

  #
  'json-formatter': {
    'website': 'https://azimi.me/json-formatter-js/',
    'register': {'alias': 'formatter', 'module': 'json-formatter.min'},
    'modules': [
      {'script': 'json-formatter.umd.min.js', 'version': '2.3.4', 'path': 'json-formatter-js@%(version)s/dist/', 'cdnjs': "https://cdn.jsdelivr.net/npm"},
    ]},

  # Datatable pivot
  'pivot': {
    'req': [{'alias': 'jqueryui'}],
    'register': {'alias': 'pivot', 'module': 'pivot.min'},
    'website': 'https://github.com/nicolaskruchten/pivottable',
    'modules': [
      {'script': 'pivot.min.js', 'version': '2.23.0', 'path': 'pivottable/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # require.js
  'require.js': {
    'website': 'https://requirejs.org/',
    'modules': [
      {'script': 'require.min.js', 'version': '2.3.6', 'path': 'require.js/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Pivot Table SubTotal
  'pivot-sub-total': {
    'req': [{'alias': 'pivot'}],
    'register': {'alias': 'subtotal', 'module': 'subtotal'},
    'website': 'http://nagarajanchinnasamy.com/subtotal/examples/',
    'modules': [
      {'script': 'subtotal.js', 'version': '1.10.0', 'path': 'subtotal@%(version)s/dist/', 'cdnjs': 'https://cdn.jsdelivr.net/npm'}]},

  # Datatable pivot C3 renderer
  'pivot-c3': {
    'req': [{'alias': 'pivot'}],
    'register': {'alias': 'pivotC', 'module': 'c3_renderers.min'},
    'website': 'https://github.com/nicolaskruchten/pivottable',
    'modules': [
      {'script': 'c3_renderers.min.js', 'version': '2.21.0', 'path': 'pivottable/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Jquery package width CDN links
  'jquery': {
    'website': 'http://jquery.com/',
    'register': {'alias': 'jQuery', 'module': 'jquery.min', 'name': '$ = jQuery', 'npm': 'jquery'},
    'modules': [
      {'script': 'jquery.min.js', 'version': '3.4.1', 'path': 'jquery/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # QUnit package width CDN links
  'qunit': {
    'website': 'https://qunitjs.com',
    'modules': [
      {'script': 'qunit.min.js', 'version': '2.9.2', 'path': 'qunit/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Used to produce sparkline charts in a document and in Tabulator
  'jquery-sparklines': {
    'req': [{'alias': 'jquery'}],
    'website': 'https://omnipotent.net/jquery.sparkline/#s-about',
    'modules': [
      {'script': 'jquery.sparkline.min.js', 'version': '2.1.2', 'path': 'jquery-sparklines/%(version)s/', 'cdnjs': CDNJS_REPO}
    ]
  },

  # Jquery UI package width CDN links
  'jqueryui': {
    'req': [{'alias': 'jquery'}],
    'website': 'http://jquery.com/',
    'register': {'alias': 'jQueryUi', 'module': 'jquery-ui.min', 'npm': 'jquery-ui'},
    'modules': [
      {'script': 'jquery-ui.min.js', 'version': '1.12.1', 'path': 'jqueryui/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Jquery-brackets package width CDN links
  'jquery-brackets': {
    'website': 'http://www.aropupu.fi/bracket/',
    'register': {'alias': 'jQueryBracket', 'module': 'jquery.bracket.min'},
    'req': [{'alias': 'jquery'}],
    'modules': [
      {'script': 'jquery.bracket.min.js', 'version': '0.11.1', 'path': 'jquery-bracket/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Jquery UI package
  # Attempt to try to solve conflict with Bootstrap
  #'jquery-ui': {'req': ['jquery', 'bootstrap'], 'modules': ['jquery-ui.min.js']},

  # Jquery timepicker width CDN links
  'timepicker': {
    'website': 'https://timepicker.co/',
    'register': {'alias': 'jQueryTime', 'module': 'jquery.timepicker.min'},
    'req': [
      {'alias': 'jquery'},
      {'alias': 'jqueryui'}],
    'modules': [
      {'script': 'jquery.timepicker.min.js', 'version': '1.3.5', 'path': 'timepicker/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # To display a context menu when right click on an item
  'jquery-context-menu': {
    'website': 'http://swisnl.github.io/jQuery-contextMenu/demo.html',
    'register': {'alias': 'jQueryContext', 'module': 'jquery.contextMenu.min'},
    'req': [{'alias': 'jquery'}, {'alias': 'jqueryui'}],
    'modules': [
      {'script': 'jquery.contextMenu.min.js', 'version': '2.6.4', 'path': 'jquery-contextmenu/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # To customize the scrollbar width CDN links
  # https://github.com/malihu/malihu-custom-scrollbar-plugin
  # http://manos.malihu.gr/repository/custom-scrollbar/demo/examples/complete_examples.html
  'jquery-scrollbar': {
    'website': 'http://manos.malihu.gr/jquery-custom-content-scroller/',
    'register': {'alias': 'jQueryScrollBar', 'module': 'jquery.mCustomScrollbar.concat.min'},
    'req': [{'alias': 'jquery'}],
    'modules': [
      {'script': 'jquery.mCustomScrollbar.concat.min.js', 'version': '3.1.5', 'path': 'malihu-custom-scrollbar-plugin/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Javascript packages for the PDF transformation width CDN links
  'pdfmake': {
    'website': '',
    'modules': [
      {'reqAlias': 'pdfmake', 'script': 'pdfmake.min.js', 'version': '0.1.57', 'path': 'pdfmake/%(version)s/', 'cdnjs': CDNJS_REPO},
      {'reqAlias': 'vfs_fonts', 'script': 'vfs_fonts.js', 'version': '0.1.57', 'path': 'pdfmake/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Javascript packages for the PDF transformation width CDN links (Tabulator)
  'jspdf': {
    'website': '',
    'modules': [
      {'reqAlias': 'jspdf', 'script': 'jspdf.min.js', 'version': '1.5.3', 'path': 'jspdf/%(version)s/', 'cdnjs': CDNJS_REPO},
      {'reqAlias': 'autotable', 'script': 'jspdf.plugin.autotable.min.js', 'version': '3.1.1', 'path': 'jspdf-autotable/%(version)s/', 'cdnjs': CDNJS_REPO},
    ]},

  # Clipboard features width CDN links
  'clipboard': {
    'website': 'https://clipboardjs.com/',
    'register': {'alias': 'clipboard', 'module': 'clipboard.min'},
    'modules': [
      {'reqAlias': 'clipboard', 'script': 'clipboard.min.js', 'version': '2.0.1', 'path': 'clipboard.js/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Javascript dependencies for D3 and NVD2 components width CDN links
  'd3': {
    'website': 'https://d3js.org/',
    'register': {'alias': 'd3', 'module': 'd3.min', 'npm': 'd3'},
    'req': [{'alias': 'jquery'}],
    'modules': [
      {'reqAlias': 'd3', 'reqMod': 'ignore', 'script': 'd3.min.js', 'version': '5.9.7', 'path': 'd3/%(version)s/', 'cdnjs': CDNJS_REPO}
    ]},

  # D3 Tips Package
  'd3-tip': {
    'req': [{'alias': 'd3'}],
    'register': {'alias': 'd3_tip', 'module': 'd3-tip.min'},
    'modules': [
      {'script': 'd3-tip.min.js', 'version': '0.9.1', 'path': 'd3-tip/%(version)s/', 'cdnjs': CDNJS_REPO}
    ]
  },

  # D3 Packages
  'd3-packages': {
      'website': 'https://github.com/d3',
      'req': [{'alias': 'd3'}],
      'modules': [
          {'reqAlias': 'd3_time_format', 'script': 'd3-time-format.min.js', 'version': '2.1.1', 'path': 'd3-time-format/%(version)s/', 'cdnjs': CDNJS_REPO},
          {'reqAlias': 'd3_shape', 'script': 'd3-shape.min.js', 'version': '1.2.0', 'path': 'd3-shape/%(version)s/', 'cdnjs': CDNJS_REPO},
          {'reqAlias': 'd3_zoom', 'script': 'd3-zoom.min.js', 'version': '1.7.1', 'path': 'd3-zoom/%(version)s/', 'cdnjs': CDNJS_REPO},
          {'reqAlias': 'd3_dsv', 'script': 'd3-dsv.min.js', 'version': '1.0.8', 'path': 'd3-dsv/%(version)s/', 'cdnjs': CDNJS_REPO},
          {'reqAlias': 'd3_ease', 'script': 'd3-ease.min.js', 'version': '1.0.3', 'path': 'd3-ease/%(version)s/', 'cdnjs': CDNJS_REPO},
          {'reqAlias': 'd3_scale', 'script': 'd3-scale.min.js', 'version': '1.0.7', 'path': 'd3-scale/%(version)s/', 'cdnjs': CDNJS_REPO},
          {'reqAlias': 'd3_brush', 'script': 'd3-brush.min.js', 'version': '1.0.4', 'path': 'd3-brush/%(version)s/', 'cdnjs': CDNJS_REPO},
          {'reqAlias': 'd3_selection', 'script': 'd3-selection.min.js', 'version': '1.2.0', 'path': 'd3-selection/%(version)s/', 'cdnjs': CDNJS_REPO},
          {'reqAlias': 'd3_transition', 'script': 'd3-transition.min.js', 'version': '1.1.1', 'path': 'd3-transition/%(version)s/', 'cdnjs': CDNJS_REPO},
          {'reqAlias': 'd3_array', 'script': 'd3-array.min.js', 'version': '1.2.2', 'path': 'd3-array/%(version)s/', 'cdnjs': CDNJS_REPO},
          {'reqAlias': 'd3_collection', 'script': 'd3-collection.min.js', 'version': '1.0.5', 'path': 'd3-collection/%(version)s/', 'cdnjs': CDNJS_REPO},
          {'reqAlias': 'd3_drag', 'script': 'd3-drag.min.js', 'version': '1.2.1', 'path': 'd3-drag/%(version)s/', 'cdnjs': CDNJS_REPO},
          {'reqAlias': 'd3_interpolate', 'script': 'd3-interpolate.min.js', 'version': '1.3.0', 'path': 'd3-interpolate/%(version)s/', 'cdnjs': CDNJS_REPO},
          {'reqAlias': 'd3_axis', 'script': 'd3-axis.min.js', 'version': '1.0.10', 'path': 'd3-axis/%(version)s/', 'cdnjs': CDNJS_REPO},
        ]},

  # D3 color module
  'd3-color': {
      'website': 'https://github.com/d3/d3-color',
      'register': {'alias': 'd3_color', 'module': 'd3-color.min'},
      'req': [{'alias': 'd3'}],
      'modules': [
        {'script': 'd3-color.min.js', 'version': '1.2.1', 'path': 'd3-color/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Javascript dependencies for Plotly width CDN links
  'plotly.js': {
    'website': 'https://plot.ly/javascript/',
    'register': {'alias': 'Plotly', 'module': 'plotly.min'},
    'req': [{'alias': 'd3'}],
    'modules': [
      {'script': 'plotly.min.js', 'version': '1.51.1', 'path': 'plotly.js/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # NVD3 Components width CDN links
  'nvd3': {
    'website': 'http://nvd3.org/',
    'req': [{'alias': 'd3', 'version': '3.5.17'}],
    'register': {'alias': 'nvd3', 'module': 'nv.d3.min'},
    'modules': [
      {'script': 'nv.d3.min.js', 'version': '1.8.6', 'path': 'nvd3/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # C3 modules width CDN links
  'c3': {
    'website': 'https://c3js.org/',
    'register': {'alias': 'c3', 'module': 'c3.min'},
    'req': [{'alias': 'd3'}],
    'modules': [
      {'script': 'c3.min.js', 'version': '0.7.15', 'path': 'c3/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  'crossfilter': {
    'website': 'http://square.github.io/crossfilter/',
    'register': {'alias': 'crossfilter', 'module': 'crossfilter.min'},
    'modules': [
      {'script': 'crossfilter.min.js', 'version': '1.3.12', 'path': 'crossfilter/%(version)s/', 'cdnjs': CDNJS_REPO}
    ]
  },

  # DC modules width CDN links
  'dc': {
    'website': 'https://dc-js.github.io/dc.js/examples/',
    'register': {'alias': 'dc', 'module': 'dc.min'},
    'req': [{'alias': 'd3'}, {'alias': 'crossfilter'},
    ],
    'modules': [
      {'script': 'dc.min.js', 'version': '3.2.1', 'path': 'dc/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # billboard modules width CDN links
  'billboard': {
    'website': 'https://naver.github.io/billboard.js/release/latest/doc/',
    'register': {'alias': 'bb', 'module': 'billboard.min'},
    'req': [{'alias': 'd3'}],
    'modules': [
      {'reqAlias': 'bb', 'script': 'billboard.min.js', 'version': '1.11.1', 'path': 'billboard.js/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # ChartJs modules width CDN links
  'Chart.js': {
    'website': 'https://www.chartjs.org/',
    'register': {'alias': 'Chart', 'module': 'Chart.min', 'npm': 'chart.js'},
    'req': [{'alias': 'd3'}],
    'modules': [
      {'script': 'Chart.bundle.min.js', 'version': '2.9.3', 'path': 'Chart.js/%(version)s/', 'cdnjs': CDNJS_REPO},
      {'script': 'Chart.min.js', 'version': '2.9.3', 'path': 'Chart.js/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # ChartJs addon to add label width CDN links
  'chartjs-pie-labels': {
      'website': 'https://chartjs-plugin-datalabels.netlify.com/',
      'register': {'alias': 'ChartDataLabel', 'module': 'chartjs-plugin-datalabels.min'},
      'req': [{'alias': 'Chart.js'}],
      'modules': [
        {'script': 'chartjs-plugin-datalabels.min.js', 'version': '0.5.0', 'path': 'chartjs-plugin-datalabels@%(version)s/dist/', 'cdnjs': 'https://cdn.jsdelivr.net/npm'}]},

  # ChartJs addon to add some Geo charts
  'chartjs-chart-geo': {
      'website': 'https://github.com/sgratzl/chartjs-chart-geo',
      'req': [{'alias': 'Chart.js'}],
      'modules': [
        {'script': 'Chart.Geo.min.js', 'version': '1.1.4', 'path': 'chartjs-chart-geo@1.1.4/build/', 'cdnjs': 'https://cdn.jsdelivr.net/npm'}]},

  # Cannot add properly the dependency in this one as my algorithm does not work for shared dependencies ....
  # 'meter': {'req': ['d3'], 'modules': ['d3.meter.js'], 'website': '', 'version': '', "status": 'deprecated'},

  # Popper tooltips used by bootstrap in the dropdown components
  'popper': {
    'req': [{'alias': 'jquery'}],
    'website': 'https://popper.js.org/',
    'register': {'alias': 'popper', 'module': 'popper.min'},
    'modules': [
      {'reqAlias': 'popper', 'script': 'popper.min.js', 'version': '1.14.6', 'path': 'popper.js/%(version)s/umd/', 'cdnjs': CDNJS_REPO}]},

  # Javascript module for the simple select component. issue with Bootstrap 4 width CDN links
  'select': {
    'website': 'http://silviomoreto.github.io/bootstrap-select/',
    'register': {'alias': 'selectBs', 'module': 'bootstrap-select.min'},
    'req': [
      {'alias': 'jquery'},
      {'alias': 'bootstrap'}],
    'modules': [
      {'reqAlias': 'selectBs', 'script': 'bootstrap-select.min.js', 'version': '1.13.9', 'path': 'bootstrap-select/%(version)s/js/', 'cdnjs': CDNJS_REPO},
    ]},

  'select-ajax': {
    'website': 'https://github.com/truckingsim/Ajax-Bootstrap-Select',
    'register': {'alias': 'selectAjax', 'module': 'ajax-bootstrap-select.min'},
    'req': [{"alias": 'select'}
    ], 'modules': [
      {'script': 'ajax-bootstrap-select.min.js', 'version': '1.4.5','path': 'ajax-bootstrap-select/%(version)s/js/', 'cdnjs': CDNJS_REPO}
    ]
  },

  # https://cdnjs.cloudflare.com/ajax/libs/ajax-bootstrap-select/1.4.5/js/ajax-bootstrap-select.min.js
  # javascript package for the Venn chart
  # 'venn': {'req': ['d3'], 'modules': ['venn.js'], 'website': '', 'version': '',},

  # Vis Javascript Packages
  'vis': {
    'website': 'http://visjs.org/',
    'register': {'alias': 'vis', 'module': 'vis.min'},
    'modules': [
      {'script': 'vis.min.js', 'version': '4.21.0', 'path': 'vis/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Javascript package to display mathematical formulas
  # https://codingislove.com/display-maths-formulas-webpage/
  # https://github.com/mathjax/mathjax
  'mathjs': {
    'website': 'https://www.mathjax.org/',
    'register': {'alias': 'mathjax', 'module': 'MathJax'},
    'package': {'zip': 'https://github.com/mathjax/MathJax/archive/%(version)s.zip', 'root': 'MathJax-%(version)s', 'folder': 'mathjax'},
    'modules': [
      {'script': 'MathJax.js', 'version': '2.7.7', 'path': 'mathjax/%(version)s/', 'cdnjs': CDNJS_REPO}],
    # To use the full module online
    #'url': 'https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.1/',
    'config': "config=TeX-AMS-MML_HTMLorMML"},

  # Socket IO
  'socket.io': {
    'website': 'https://github.com/socketio/socket.io',
    'register': {'alias': 'socketio', 'module': 'socket.io'},
    'req': [{'alias': 'jquery'}],
    'modules': [
      {'script': 'socket.io.js', 'version': '2.2.0', 'path': 'socket.io/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Code mirror
  'codemirror': {
    'website': 'https://codemirror.net/',
    'modules': [
      {'script': 'codemirror.min.js', 'version': '5.52.2', 'path': 'codemirror/%(version)s/', 'cdnjs': CDNJS_REPO}
    ]
  },

  # Leaflet
  'leaflet': {
    'website': 'https://leafletjs.com/',
    'modules': [
      {'script': 'leaflet.js', 'version': '1.6.0', 'path': 'leaflet/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # showdown
  'showdown': {
    'website': 'https://github.com/showdownjs/showdown',
    'modules': [
      {'script': 'showdown.min.js', 'path': 'showdown/%(version)s/', 'version': '1.9.1', 'cdnjs': CDNJS_REPO}
    ]
  }
}


CSS_IMPORTS = {
  'jqueryui': {
    'website': 'http://jquery.com/',
    'modules': [
      {'script': 'jquery-ui.min.css', 'version': '1.12.1', 'path': 'jqueryui/%(version)s/', 'cdnjs': CDNJS_REPO},
    ]
  },

  # QUnit package width CDN links
  'qunit': {
    'website': 'https://qunitjs.com',
    'modules': [
      {'script': 'qunit.min.css', 'version': '2.9.2', 'path': 'qunit/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Jquery-brackets package width CDN links
  'jquery-brackets': {
    'req': [{'alias': 'jqueryui'}],
    'modules': [
      {'script': 'jquery.bracket.min.css', 'version': '0.11.1', 'path': 'jquery-bracket/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # To display a context menu when right click on an item width CDN links
  # http://swisnl.github.io/jQuery-contextMenu/demo.html#jquery-context-menu-demo-gallery
  'jquery-context-menu': {
    'website': 'https://github.com/swisnl/jQuery-contextMenu/blob/master/dist/jquery.contextMenu.min.css.map',
    'req': [{'alias': 'jqueryui'}],
    'modules': [
      {'script': 'jquery.contextMenu.min.css', 'version': '2.6.4', 'path': 'jquery-contextmenu/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Jquery timepicker width CDN links
  'timepicker': {
    'website': 'https://timepicker.co/',
    'req': [{'alias': 'jqueryui'}],
    'modules': [
      {'script': 'jquery.timepicker.min.css', 'version': '1.3.5', 'path': 'timepicker/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # To customize the scrollbar width CDN links
  'jquery-scrollbar': {
    'website': 'http://manos.malihu.gr/jquery-custom-content-scroller/',
    'req': [{'alias': 'jqueryui'}],
    'modules': [
      {'script': 'jquery.mCustomScrollbar.min.css', 'version': '3.1.5', 'path': 'malihu-custom-scrollbar-plugin/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Tabulator definition
  'tabulator': {
    'website': 'http://tabulator.info',
    'modules': [
      {'script': 'tabulator.min.css', 'version': '4.4.3', 'path': 'tabulator/%(version)s/css/', 'cdnjs': CDNJS_REPO}
    ]
  },

  'datatables': {
    'website': 'https://datatables.net/',
    'req': [{'alias': 'bootstrap'}],
    'modules': [
      {'script': 'jquery.dataTables.min.css', 'version': '1.10.19', 'path': '%(version)s/css/', 'cdnjs': 'https://cdn.datatables.net'}
  ]},

  # Datatable Buttons
  'datatables-buttons': {
    'website': 'https://datatables.net/extensions/buttons/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'buttons.dataTables.min.css', 'version': '1.6.1', 'path': 'buttons/%(version)s/css/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable Select
  'datatables-select': {
    'website': 'https://datatables.net/extensions/select/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'select.dataTables.min.css', 'version': '1.3.1', 'path': 'select/%(version)s/css/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable Scroller
  'datatables-scroller': {
    'website': 'https://datatables.net/extensions/scroller/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'scroller.dataTables.min.css', 'version': '2.0.1', 'path': 'scroller/%(version)s/css/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable SearchPanes
  'datatables-searchPanes': {
    'website': 'https://datatables.net/extensions/searchpanes/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'searchPanes.dataTables.min.css', 'version': '1.0.1', 'path': 'searchpanes/%(version)s/css/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable Responsive
  'datatables-responsive': {
    'website': 'https://datatables.net/extensions/responsive/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'responsive.dataTables.min.css', 'version': '2.2.3', 'path': 'responsive/%(version)s/css/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable KeyTable
  'datatables-keytable': {
    'website': 'https://datatables.net/extensions/keytable/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'keyTable.dataTables.min.css', 'version': '2.5.1', 'path': 'keytable/%(version)s/css/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable autoFill
  'datatables-autoFill': {
    'website': 'https://datatables.net/extensions/autofill/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'autoFill.dataTables.min.css', 'version': '2.1.0', 'path': 'autofill/%(version)s/css/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable group row
  'datatables-row-group': {
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'rowGroup.dataTables.min.css', 'version': '1.1.1', 'path': 'rowgroup/%(version)s/css/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable fixed column
  'datatables-fixed-columns': {
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'fixedColumns.bootstrap4.min.css', 'version': '3.2.2', 'path': 'fixedcolumns/%(version)s/css/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable fixed header
  'datatables-fixed-header': {
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'fixedHeader.bootstrap4.min.css', 'version': '3.1.3', 'path': 'fixedheader/%(version)s/css/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable export module
  'datatables-export': {
    'website': 'https://datatables.net/',
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'buttons.bootstrap4.min.css', 'version': '1.5.2', 'path': 'buttons/%(version)s/css/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Datatable column ordering
  'datatables-col-order': {
    'req': [{'alias': 'datatables'}],
    'modules': [
      {'script': 'colReorder.bootstrap4.min.css', 'version': '1.5.1', 'path': 'colreorder/%(version)s/css/', 'cdnjs': 'https://cdn.datatables.net'}]},

  # Bootstrap style width CDN links
  'bootstrap': {
    'website': 'https://getbootstrap.com/',
    'req': [{'alias': 'font-awesome'}],
    'modules': [
      {'script': 'bootstrap.min.css', 'version': '4.4.1', 'path': 'bootstrap/%(version)s/css/', 'cdnjs': 'https://stackpath.bootstrapcdn.com'}]},

  # Font awesome style width CDN links
  'font-awesome': {
    'website': 'https://fontawesome.com/',
    'package': {'zip': 'https://use.fontawesome.com/releases/v%(version)s/fontawesome-free-%(version)s-web.zip',
                    'root': 'fontawesome-free-%(version)s-web', 'folder': 'releases', 'path': 'v%(version)s'},
    'modules': [
      {'script': 'all.css', 'version': '5.11.2', 'path': 'releases/v%(version)s/css/', 'cdnjs': 'https://use.fontawesome.com'}]},

  # NVD3 Components width CDN links
  'nvd3': {
    'website': 'http://nvd3.org/',
    'modules': [
      {'script': 'nv.d3.min.css', 'version': '1.8.6', 'path': 'nvd3/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # C3 modules width CDN links
  'c3': {
    'website': 'https://c3js.org/',
    'modules': [
      {'script': 'c3.min.css', 'version': '0.7.15', 'path': 'c3/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # DC modules width CDN links
  'dc': {
    'website': 'https://dc-js.github.io/dc.js/examples/',
    'modules': [
      {'script': 'dc.min.css', 'version': '3.2.1', 'path': 'dc/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # billboard modules width CDN links
  'billboard': {
    'modules': [
      {'script': 'billboard.min.css', 'version': '1.11.1', 'path': 'billboard.js/%(version)s/', 'cdnjs': CDNJS_REPO}],
    'website': 'https://naver.github.io/billboard.js/release/latest/doc/'},

  #'epyk': {'req': ['bootstrap'], 'modules': ['bdi.css'], 'website': 'internal lib', 'version': '0'},

  # Javascript module for the simple select component. issue with Bootstrap 4 width CDN links
  'select': {
    'website': 'https://github.com/silviomoreto/bootstrap-select',
    'req': [{'alias': 'jqueryui'}, {'alias': 'bootstrap'}],
    'modules': [
      {'script': 'bootstrap-select.min.css', 'version': '1.13.6', 'path': 'bootstrap-select/%(version)s/css/', 'cdnjs': CDNJS_REPO}
    ]},

  'select-ajax': {
    'website': 'https://github.com/truckingsim/Ajax-Bootstrap-Select',
    'req': [{"alias": 'select'}],
    'modules': [
          {'script': 'ajax-bootstrap-select.min.css', 'version': '1.4.5', 'path': 'ajax-bootstrap-select/%(version)s/css/', 'cdnjs': CDNJS_REPO}
    ]
  },

  # Pivot table style with CDN Links
  'pivot': {
    'website': 'https://github.com/nicolaskruchten/pivottable',
    'modules': [
      {'script': 'pivot.min.css', 'version': '2.23.0', 'path': 'pivottable/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Vis style with CDN Links
  'vis': {
    'website': 'http://visjs.org/',
    'modules': [
      {'script': 'vis.min.css', 'version': '4.21.0', 'path': 'vis/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  # Code mirror
  'codemirror': {
    'website': 'https://codemirror.net/',
    'modules': [
      {'script': 'codemirror.min.css', 'version': '5.52.2', 'path': 'codemirror/%(version)s/', 'cdnjs': CDNJS_REPO}
    ]},

  # Leaflet
  'leaflet': {
    'website': 'https://leafletjs.com/',
    'modules': [
      {'script': 'leaflet.css', 'version': '1.6.0', 'path': 'leaflet/%(version)s/', 'cdnjs': CDNJS_REPO}]},

  #
  'json-formatter': {
    'website': 'https://azimi.me/json-formatter-js/',
    'modules': [
      {'script': 'json-formatter.css', 'version': '2.3.4', 'path': 'json-formatter-js@%(version)s/dist/', 'cdnjs': "https://cdn.jsdelivr.net/npm"},
    ]},
}


_SERVICES = {}


def extend(reference, module_path, version, cdnjs_url=CDNJS_REPO, required=None):
  """
  Description:
  ------------
  Function to extend the internal CSS and JS registered modules.

  Related Pages:
  --------------
  :param reference: String. The internal reference in the framework
  :param module_path: List of tuple. The different modules and location
  :param version: String. The version number. Can be an internal module reference to point to follow its version number
  :param cdnjs_url: String. The CDNJS reference path
  :param required: List. The list of dependency modules
  """
  for module, path in module_path:
    config = JS_IMPORTS if module.endswith(".js") else CSS_IMPORTS
    if not reference in config:
      config[reference] = {"modules": []}
      if required is not None:
        reqs = [{'alias': req} for req in required if req in config]
        if reqs:
          config[reference]['req'] = reqs
    if version in config:
      # take the version from another registered module
      version = config[version]['modules'][0]['version']
    config[reference]["modules"].append({'script': module, 'version': version, 'path': path, 'cdnjs': cdnjs_url})


def extend_imports(extension):
  """
  Description:
  ------------
  Hook to extend the imports in the centralised Import module.
  The packages definition is quite similar to the one in IMports.py except that CSS and JS are grouped together for
  simplicity

  :param extension: Dictionary. The list of packages to be added grouped by alias
  """
  global CSS_IMPORTS, JS_IMPORTS

  for alias, mod in extension.items():
    css, js = {'website': mod.get('website', ''), 'modules': []}, {'website': mod.get('website', ''), 'modules': []}
    if 'register' in mod:
      js['register'] = mod['register']
    if 'req' in mod:
      css['req'], js['req'] = [], []
      for req in mod['req']:
        if req['alias'] in CSS_IMPORTS:
          css['req'].append(req)
        if req['alias'] in JS_IMPORTS:
          js['req'].append(req)
    if 'modules' in mod:
      for module in mod['modules']:
        if not 'cdnjs' in module:
          module['cdnjs'] = CDNJS_REPO
        module['version'] = mod['version']  # propagate the version tag
        if module['script'].endswith(".js"):
          js['modules'].append(module)
        elif module['script'].endswith(".css"):
          css['modules'].append(module)
      if css['modules']:
        CSS_IMPORTS[alias] = css
      if js['modules']:
        JS_IMPORTS[alias] = js
    if 'services' in mod:
      for service in mod['services']:
        service['pmts'] = ";".join(["%s=%s" % (k, v) for k, v in service['values'].items()])
        if service['type'] not in _SERVICES:
          _SERVICES[alias] = {}
        _SERVICES[alias].setdefault(service['type'], []).append("%(url)s?%(pmts)s" % service)


class ImportManager(object):
  """
  Description:
  ------------
  The main class in charge of defining the order of the imports in the header.

  There is no check on the presence of the modules on the server. The only purpose of this module is to produce the
  string with the module names and the correct paths to your final HTML report.
  """

  def __init__(self, online=False, report=None):
    """
    Description:
    ------------
    Load the hierarchy of modules.
    This module will define the import section in the header of the final HTML page.

    It will create links to the official online websites or link to an internal copy if no internet connection is
    available. To run a report using the online mode to False it is requires to get all the packages locally
    saved with the expected structured (basically the one of the CDNJ repository)

    Attributes:
    ----------
    :param online: Optional. A flag to specify if the report can use an internet connection. Default False
    :param report: Optional. The internal report object with all the required external modules
    """
    self._report, ovr_version = report, {}
    if report is not None and report.ext_packages is not None:
      extend_imports(report.ext_packages)
    if report is not None and self._report.run.report_name is not None and self._report.run.local_path is not None and os.path.exists(os.path.join(self._report.run.local_path, '__init__.py')):
      # Force the version of some external Javascript or CSS packages
      packages = importlib.import_module("%s.__init__" % self._report.run.report_name)
      ovr_version = getattr(packages, 'MODULES', {})
    if report is not None:
      # Apply the different reports overrides on the packages versions
      ovr_version.update(self._report._props.get('packages', {}))
    self.jsImports, self.cssImports, self.moduleConfigs, self.reqVersion = {}, {}, {}, {}
    for folder, import_cict, import_type in [('js', self.jsImports, JS_IMPORTS), ('css', self.cssImports, CSS_IMPORTS)]:
      for alias, definition in import_type.items():
        main = collections.OrderedDict()
        for i, mod in enumerate(definition['modules']):
          if alias in ovr_version:
            mod['version'] = ovr_version[alias]
          script = "".join([mod['path'] % mod, mod['script']])
          if online:
            main["%s/%s" % (mod['cdnjs'], script)] = mod['version']
          elif 'url' in definition:
            main["%s%s" % (definition['url'], script)] = mod['version']
          else:
            main["%s/%s" % (STATIC_PATH.replace("\\", "/"), script)] = mod['version']
        modules = collections.OrderedDict()
        self.getModules(modules, alias, folder, import_type)
        if 'config' in definition:
          self.moduleConfigs[alias] = definition['config']

        mainKeys, versions = [], []
        for k, v in main.items():
          mainKeys.append(k)
          versions.append(v)
        import_cict[alias] = {'main': main, 'dep': list(modules.keys()), 'versions': versions}

  def getModules(self, modules, alias, folder=None, module_details=None):
    """
    Description:
    ------------
    Return the list of modules for a given entry.
    This will be used recursively to resolve all the dependencies

    Usage::

      modules = collections.OrderedDict()
    ImportManager().getModules(modules, 'c3')

    Attributes:
    ----------
    :param modules: The list of modules
    :param alias: The module reference in the above JS and CSS dictionaries
    :param folder: Optional. The folder name
    :param module_details: Optional. The module definition. Default check in the Javascipt modules

    :return: The list of modules
    """
    if module_details is None:
      module_details = dict(JS_IMPORTS)
    if isinstance(alias, dict):
      alias = alias['alias']

    for mod in module_details[alias]['modules']:
      script = "".join([mod['path'] % mod, mod['script']])
      if 'url' in module_details[alias]:
        modules["%s/%s" % (module_details[alias]['url'], script)] = True
      else:
        modules[r"%s\%s" % (STATIC_PATH.replace("\\", "/"), script)] = True
    for req in module_details.get(alias, {}).get('req', []):
      self.getModules(modules, req, folder, module_details)
    return modules

  def getReq(self, mod, modules, import_hierarchy):
    """
    Description:
    ------------
    Returns the list pf required modules for a given alias

    Attributes:
    ----------
    :param mod:
    :param modules:
    :param import_hierarchy:

    """
    if isinstance(mod, dict):
      # This will allow different versions of packages according to the modules
      # For example NVD3 cannot use any recent version of D3
      if 'version' in mod:
        self.reqVersion[mod['alias']] = mod['version']
        new_main_for_alias, new_main_for_alias_css = collections.OrderedDict(), collections.OrderedDict()
        for path in self.jsImports[mod['alias']]['main']:
          for v in self.jsImports[mod['alias']]['versions']:
            new_main_for_alias[path.replace(v, mod['version'])] = mod['version']
        if mod['alias'] in self.cssImports:
          for path in self.cssImports[mod['alias']]['main']:
            for v in self.cssImports[mod['alias']]['versions']:
              new_main_for_alias_css[path.replace(v, mod['version'])] = mod['version']
          self.cssImports[mod['alias']]['main'] = new_main_for_alias_css
        # Store the new dictionary with the key and version updated for the module
        self.jsImports[mod['alias']]['main'] = new_main_for_alias
        for i, path in enumerate(self.jsImports[mod['alias']]['dep']):
          for v in self.jsImports[mod['alias']]['versions']:
            path = path.replace(v, mod['version'])
          self.jsImports[mod['alias']]['dep'][i] = path
      mod = mod['alias']
    modules.append(mod)
    for req in import_hierarchy.get(mod, {}).get("req", []):
      self.getReq(req, modules, import_hierarchy)

  def cleanImports(self, imports, import_hierarchy):
    """
    Description:
    ------------
    Remove the underlying imports to avoid duplicated entries

    Usage::

      >>> ImportManager().cleanImports(['c3'], JS_IMPORTS)
    ['jquery', 'd3', 'c3']

    Attributes:
    ----------
    :param imports: An array with the list of aliases for the external packages
    :param import_hierarchy: The package definition (Javascript or CSS) from the above import lists

    :return: Return the list with the full list of aliases (including dependencies)
    """
    import_resolved = []
    for mod in imports:
      self.getReq(mod, import_resolved, import_hierarchy)
    for a in set(import_resolved):
      occurences = [j for j, x in enumerate(import_resolved) if x == a]
      if len(occurences) > 1:
        for j in occurences[::-1][1:]:
          import_resolved.pop(j)
    return import_resolved[::-1]

  def cssResolve(self, css_aliases, local_css=None, excluded=None):
    """
    Description:
    ------------
    Return the list of CSS modules to add to the header

    Usage::

      >>> ImportManager().cssResolve(['c3'])
    '<link rel="stylesheet" href="/static/c3/0.6.12/c3.min.css" type="text/css">'

    Attributes:
    ----------
    :param css_aliases: An array with the list of aliases for the external packages
    :param local_css: Optional. The file overrides
    :param excluded: Optional. Packages excluded from the result object (mandatory for some frameworks already emboarding modules)

    :return: The string to be added to the header
    """
    css = []
    css_aliases = self.cleanImports(css_aliases, CSS_IMPORTS)
    for css_alias in css_aliases:
      if excluded is not None and css_alias in excluded:
        continue

      if css_alias in _SERVICES:
        # Add services url
        for service in _SERVICES[css_alias].get('css', []):
          css.append('<link rel="stylesheet" href="%s">' % service)
        continue

      for urlModule in list(self.cssImports[css_alias]['main']):
        css.append('<link rel="stylesheet" href="%s" type="text/css">' % urlModule)
    if local_css is not None:
      for localCssFile in local_css:
        css.append('<link rel="stylesheet" href="%s" type="text/css">' % localCssFile)
    return "\n".join(css)

  def cssURLs(self, css_str):
    """
    Description:
    ------------
    Retrieve the list of CSS dependencies URL from a header

    Attributes:
    ----------
    :param css_str: The CSS String in the page

    :return: A Python list with all the CSS external URL to be imported
    """
    return re.findall('<link rel="stylesheet" href="(.*?)" type="text/css">', css_str)

  def jsResolve(self, js_aliases, local_js=None, excluded=None):
    """
    Description:
    ------------
    Return the list of Javascript modules to add to the header

    Usage::

      >>> ImportManager().jsResolve(['c3'])
    '<script language="javascript" type="text/javascript" src="/static/jquery/3.4.1/jquery.min.js"></script>\n<script language="javascript" type="text/javascript" src="/static/d3/5.9.7/d3.min.js"></script>\n<script language="javascript" type="text/javascript" src="/static/c3/0.6.12/c3.min.js"></script>'

    Attributes:
    ----------
    :param js_aliases: An array with the list of aliases for the external packages
    :param local_js: Optional. The file overrides
    :param excluded: Optional. Packages excluded from the result object (mandatory for some frameworks already emboarding modules)

    :return: The string to be added to the header
    """
    js = []
    js_aliases = self.cleanImports(js_aliases, JS_IMPORTS)
    for js_alias in js_aliases:
      if excluded is not None and js_alias in excluded:
        continue

      extra_configs = "?%s" % self.moduleConfigs[js_alias] if js_alias in self.moduleConfigs else ""
      for url_module in list(self.jsImports[js_alias]['main']):
        if '/mode/' in url_module:
          js.append('<script type="module" language="javascript" src="%s%s"></script>' % (url_module, extra_configs))
        else:
          js.append('<script language="javascript" type="text/javascript" src="%s%s"></script>' % (url_module, extra_configs))
    if local_js is not None and len(local_js) > 0:
      for local_js_file in local_js:
        js.append('<script language="javascript" type="text/javascript" src="%s"></script>' % local_js_file)
    return "\n".join(js)

  def jsURLs(self, js_str):
    """
    Description:
    ------------
    Retrieve the list of Javascript dependencies URL from a header

    Attributes:
    ----------
    :param js_str: The Javascript String in the page

    :return: A Python list with all the Javascript external URL to be imported
    """
    return re.findall('<script language="javascript" type="text/javascript" src="(.*?)"></script>', js_str)

  def getFiles(self, cssAlias, jsAlias):
    """
    Description:
    ------------
    retrieve the package definition from the list of module aliases

    Usage::

      >>> ImportManager().getFiles(['c3'], ['c3'])
    f['css'][0]['file']['script']

    Attributes:
    ----------
    :param cssAlias: An array with the list of aliases for the CSS external packages
    :param jsAlias: An array with the list of aliases for the Js external packages

    :return: A dictionary with the CSS and JS files definition
    """
    files = {'css': [], 'js': []}
    mod_css, mod_js = {}, {}
    for alias, details in CSS_IMPORTS.items():
      mod_css[alias] = []
      for module in details['modules']:
        mod_css[alias].append({'version': module.get('version', ''), 'alias': alias, 'file': module, 'website':
          details.get('website', ''), 'status': details.get('status', '')})
    for alias, details in JS_IMPORTS.items():
      mod_js[alias] = []
      for module in details['modules']:
        mod_js[alias].append({'version': module.get('version', ''), 'alias': alias, 'file': module, 'website':
          details.get('website', ''), 'status': details.get('status', '')})
    for css_file in self.cleanImports(cssAlias, CSS_IMPORTS):
      files['css'].extend(mod_css[css_file])
    for js_file in self.cleanImports(jsAlias, JS_IMPORTS):
      files['js'].extend(mod_js[js_file])
    return files

  def cssGetAll(self):
    """
    Description:
    ------------
    To retrieve the full list of available modules on the server
    """
    return self.cssResolve(set(CSS_IMPORTS.keys()))

  def jsGetAll(self):
    """
    Description:
    ------------
    To retrieve the full list of available modules on the server
    """
    return self.jsResolve(set(JS_IMPORTS.keys()))

  def getPackage(self, alias, version=None, static_path=None, with_dep=False, reload=True):
    """
    Description:
    ------------
    Function in charge of downloading the different external CSS and JS packages locally.
    This will guarantee the install without having to get any extra features saved on a repository.
    Saved copies of the modules can be done in order to guarantee a offline mode

    Usage::

      Imports.ImportManager(report=Report()).getPackage('jqueryui')

    Attributes:
    ----------
    :param alias: The package reference in the above list
    :param version: Optional. The package version to retrieve
    :param static_path: Optional. The path in which the files should be copied to
    :param with_dep: Optional. Flag to specify if the dependencies should be updated. Default False
    :param reload: Optional. Flag to force the package reloading if the folder already exists. Default Yes

    """
    if not hasattr(self._report, "py"):
      from epyk.core.py.PyRest import PyRest
      webscrapper = PyRest().webscrapping
    else:
      webscrapper = self._report.py.requests.webscrapping

    if not static_path.endswith("static"):
      static_path = os.path.join(static_path, "static")
    packages = {}
    _static_path = os.path.join(os.path.dirname(__file__), '..', '..', 'static') if static_path is None else static_path
    if not _static_path.endswith("static"):
      _static_path = os.path.join(_static_path, "static")
    for pckg in [JS_IMPORTS, CSS_IMPORTS]:
      if with_dep:
        for depAlias in self.cleanImports([alias], pckg):
          if depAlias != alias:
            self.getPackage(depAlias, reload=reload)
      if 'package' in pckg.get(alias, {}):
        packages[alias] = os.path.join(_static_path, pckg[alias]['package']['folder'])
      for mod in pckg.get(alias, {}).get('modules', []):
        _version = self.reqVersion.get(alias, mod['version']) if version is None else version
        script = "".join([mod['path'] % {'version': _version}, mod['script']])
        path = os.path.join(_static_path, mod['path'] % {'version': _version})
        if not os.path.exists(path):
          os.makedirs(path)
        reloadModule = True
        extFilePath = r"%s\%s" % (path, mod['script'])
        if os.path.exists(extFilePath) and not reload:
          reloadModule = False

        if reloadModule:
          page = webscrapper("%s/%s" % (mod['cdnjs'], script))
          if hasattr(page, 'code') and page.code == 404:
            print(" # Error - %s: Script %s/%s not found " % (alias, mod['cdnjs'], script))
            continue

          try:
            extFileName = open(extFilePath, "wb")
            extFileName.write(page)
            extFileName.close()
            print("  > %s - %s, version %s. Done !" % (alias, mod['script'], _version))
          except Exception as err:
            print(" # Exception - %s: %s/%s, %s" % (alias, mod['script'], _version, err))
            print(err)
        else:
          print("  > %s - %s, version %s. Already defined !" % (alias, mod['script'], _version))

    if len(packages) > 0:
      print("")
      print("Downloading %s packages, this might take few minutes" % len(packages))
      for pckg, folder in packages.items():
        self.getFullPackage(pckg, version=version, static_path=static_path, reload=reload)

  def getFullPackage(self, alias, version=None, static_path=None, reload=False):
    """
    Description:
    ------------
    Download a full package (CSS and JS) locally for a server or full offline mode

    Usage::

      Imports.ImportManager(report=Report()).getFullPackage('font-awesome')

    Attributes:
    ----------
    :param alias: The package reference in the above list
    :param version: Optional. The package version to retrieve
    :param static_path: Optional. The path in which the files should be copied to
    :param reload: Optional. Flag to force the package reloading if the folder already exists. Default False

    :return: The Python Import manager
    """

    import zipfile
    import shutil
    import io
    import os

    if not hasattr(self._report, "py"):
      from epyk.core.py.PyRest import PyRest
      webscrapper = PyRest().webscrapping
    else:
      webscrapper = self._report.py.requests.webscrapping

    if 'package' in JS_IMPORTS[alias]:
      versionDict = {'version': JS_IMPORTS[alias]['modules'][0]['version'] if version is None else version}
      packagePath = JS_IMPORTS[alias]['package']['zip'] % versionDict
      if static_path is None:
        static_path = os.path.join(os.path.dirname(__file__), '..', '..', 'static', JS_IMPORTS[alias]['package']['folder'])
      else:
        static_path = os.path.join(static_path, "static")
      if not os.path.exists(static_path):
        # Create the destination folders if missing
        os.makedirs(static_path)
      dstPath = os.path.join(static_path, JS_IMPORTS[alias]['package'].get('folder', ''), JS_IMPORTS[alias]['package'].get('path', '%(version)s') % versionDict)
      vReloadPath = True
      if os.path.exists(dstPath):
        if not reload:
          vReloadPath = False
        else:
          shutil.rmtree(dstPath)

      if vReloadPath:
        print("  > Downloading package %s" % packagePath)
        r = webscrapper(packagePath)
        z = zipfile.ZipFile(io.BytesIO(r))
        #z = zipfile.ZipFile(io.BytesIO(r))
        z.extractall(static_path)
        if JS_IMPORTS[alias]['package']['root'] is not None:
          root = JS_IMPORTS[alias]['package']['root'] % versionDict
          shutil.copytree(os.path.join(static_path, root), dstPath)
          shutil.rmtree(os.path.join(static_path, root))
        print("  < Package %s. Done ! " % alias)
      else:
        print("  < Package %s already loaded " % alias)
    return self

  def package(self, alias):
    """
    Description:
    ------------
    Returns the packages used in the Framework for both Js and CSS perimeters

    Usage::

      >>> len(ImportManager().package('jqueryui')['modules'])
    4

    Attributes:
    ----------
    :param alias: The package reference in the above lists

    :return: A dictionary with the package details
    """
    res = {}
    if alias in CSS_IMPORTS:
      res.update(CSS_IMPORTS[alias])
    if alias in JS_IMPORTS:
      for k, v in JS_IMPORTS[alias].items():
        if k in res:
          if isinstance(v, list):
            res[k].extend(v)
        else:
          res[k] = v
    return res

  def setVersion(self, alias, version):
    """
    Description:
    ------------
    Allow the use of different version of a package.

    This will change the Import important to the Python env

    Attributes:
    ----------
    :param alias: The package reference in the above list
    :param version: The new version to be used globally

    """
    self.reqVersion[alias] = version
    for modType in [CSS_IMPORTS, JS_IMPORTS]:
      if alias in modType:
        for mod in modType[alias].get('modules', []):
          mod['version'] = version

  def addPackage(self, alias, config):
    """
    Description:
    ------------
    Add a new package or update an existing one with new parameters.
    Only few parameters are available here in order to limit the changes

    Usage::

      i.addPackage('test',
      {
        'req': [{'alias': 'd3'}],
        'modules': [
          {'script': 'dc.min.css', 'version': '3.0.9', 'path': 'dc/%(version)s/', 'cdnjs': 'https://cdnjs.cloudflare.com/ajax/libs'},
          {'script': 'dc.min.js', 'version': '3.0.9', 'path': 'dc/%(version)s/', 'cdnjs': 'https://cdnjs.cloudflare.com/ajax/libs'},
        ]},
      )

    Attributes:
    ----------
    :param alias: The package alias
    :param config: The Python dictionary with the package details

    :return: The import Manager
    """

    global CSS_IMPORTS
    global JS_IMPORTS

    mod_entry = {'css': {}, 'js': {}}
    for mod in config['modules']:
      if mod['script'].endswith(".css"):
        mod_entry['css'].setdefault('modules', []).append(mod)
        if 'req' in config:
          for req in config['req']:
            if req['alias'] in CSS_IMPORTS:
              mod_entry['css'].setdefault('req', []).append(req)
      elif mod['script'].endswith(".js"):
        mod_entry['js'].setdefault('modules', []).append(mod)
        if 'req' in config:
          for req in config['req']:
            if req['alias'] in JS_IMPORTS:
              mod_entry['js'].setdefault('req', []).append(req)
    if len(mod_entry['css']) > 0:
      CSS_IMPORTS.setdefault(alias, {}).update(mod_entry['css'])
    if len(mod_entry['js']) > 0:
      JS_IMPORTS.setdefault(alias, {}).update(mod_entry['js'])
    return self

  def getPackages(self, static_path=None, reload=False, exclude=None):
    """
    Description:
    ------------
    Download all the CSS and Js packages from the official CDNJS configured in the configuration.
    It is possible to get the configuration settings by calling the function getPackageInfo(aliasName) attached to the report

    Attributes:
    ----------
    :param static_path: The package reference in the above list
    :param reload: Optional. Flag to force the package reloading if the folder already exists. Default False

    :return: The Python Import manager
    """
    if exclude is None:
      exclude = []
    if not static_path.endswith("static"):
      static_path = os.path.join(static_path, "static")
    aliases = list(set(list(CSS_IMPORTS.keys()) + list(JS_IMPORTS.keys())))
    for alias in aliases:
      self.getPackage(alias, static_path=static_path, reload=reload)
    return self

  def to_requireJs(self, data, excluded_packages=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data: Dictionary. The Report modules to resolve
    :param excluded_packages: Optional. List. The packages to exclude
    """
    deps_level, alias_to_name, name_to_alias, results = {}, {}, {}, {'jsFrgs': data['jsFrgs'], 'paths': {}}
    m_versions = {}

    # Check first if some specific versions are requried for the packages
    for m in self._report.jsImports:
      for req in JS_IMPORTS[m].get('req', []):
        if 'version' in req:
          m_versions[req['alias']] = req['version']

    # Produce the dependency tree for requirejs
    for m in self.cleanImports(self._report.jsImports, JS_IMPORTS):
      if excluded_packages is not None and m in excluded_packages:
        continue

      if 'register' in JS_IMPORTS[m]:
        alias = JS_IMPORTS[m]['register']['alias']
        first_module = JS_IMPORTS[m]['modules'][0]
        if m in m_versions:
          first_module['version'] = m_versions[m]
        results['paths'][alias] = "%s/%s%s" % (first_module['cdnjs'], first_module['path'] % first_module,
                                     JS_IMPORTS[m]['register'].get('module', first_module['script'][:-3]))
        alias_to_name[m] = alias
        name_to_alias[alias] = m
        if 'req' in JS_IMPORTS[m]:
          max_level = max([deps_level[alias['alias']] if alias['alias'] in deps_level else -1 for alias in
                           JS_IMPORTS[m]['req']])
          deps_level[m] = max_level + 1
        else:
          deps_level[m] = 0

    # Create the chains of modules to be loaded
    level, group = None, []
    for k, v in sorted(deps_level.items(), key=lambda item: item[1])[::-1]:
      if level is None:
        level = v
      if level != v:
        for g in group:
          if 'init_fnc' in JS_IMPORTS[name_to_alias[g]]['register']:
            results['jsFrgs'] = "%s; %s" % (JS_IMPORTS[name_to_alias[g]]['register']['init_fnc'], results['jsFrgs'])
        results['jsFrgs'] = "require(['%s'], function (%s) { %s })" % (
        "', '".join([g for g in group]), ", ".join([g for g in group]), results['jsFrgs'])
        level, group = v, [alias_to_name[k]]
      else:
        group.append(alias_to_name[k])
    if group:
      for g in group:
        if 'init_fnc' in JS_IMPORTS[name_to_alias[g]]['register']:
          results['jsFrgs'] = "%s; %s" % (JS_IMPORTS[name_to_alias[g]]['register']['init_fnc'], results['jsFrgs'])
      results['jsFrgs'] = "require(['%s'], function (%s) { %s })" % (
      "', '".join([g for g in group]), ", ".join([g for g in group]), results['jsFrgs'])
    return results




from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObject


class JsWhile(object):

  def __init__(self, jsCondition, ruleType='boolean', iterVar="i", start=0, step=1, context=None):
    """
    Marks a block of statements to be executed while a condition is true

    Documentation:
      - https://www.w3schools.com/jsref/jsref_while.asp

    :param jsCondition:
    :param ruleType:
    :param iterVar:
    :param start:
    :param step:
    :param context:
    """
    self._context, self.__whileId, self.__iterVar, self.__next = context, None, None, None
    self.__whileDataId = None
    if ruleType == 'array':
      self.step = step
      self.__jsObj = JsObject.JsObject(jsCondition, setVar=True)
      self.__iterVar = iterVar
      self.__val = JsObject.JsObject("{key: %(keys)s, value: %(dico)s[%(keys)s]}" % {"dico": self.__jsObj, "keys": iterVar}, isPyData=False, setVar=True)
      self.next("%s+=%s" % (self.__iterVar, self.step))
      # Should add undefined in order not to stop in the zero values in the iterator
      self.__while = ["var %s = %s" % (self.__jsObj, jsCondition), "var %s = %s" % (self.__iterVar, start),
                      "while(%s[i] !== undefined)" % self.__jsObj.varName]
    elif ruleType == 'dict':
      self.__jsObj = JsObject.JsObject(jsCondition, setVar=True)
      self.__iterVar = "%s_keys" % self.__jsObj.varName
      self.__val = JsObject.JsObject("{key: %(keys)s[k], value: %(dico)s[%(keys)s[k]]}" % {"dico": self.__jsObj, "keys": self.__iterVar}, isPyData=False, setVar=True)
      self.next("k += 1")
      # Should add undefined in order not to stop in the zero values in the iterator
      self.__while = ["var %s = %s" % (self.__jsObj, jsCondition), "var %s = %s" % (self.__iterVar, self.__jsObj.keys()),
                      "var k = 0", "while(%s[k] !== undefined)" %  self.__iterVar]
    else:
      self.__while = ["while(%s)" % jsCondition]
    self.__jsFncs = []

  def next(self, rule):
    """

    :return:
    """
    self.__next = rule
    return self

  def js(self, jsFncs, isPyData=False, jsFncVal=None):
    """

    :param jsFncs:
    :param isPyData:
    :param jsFncVal:
    :return:
    """
    if jsFncVal is None:
      jsFncVal = self.__val.varName
    jsIterable = JsUtils.jsConvertFncs(jsFncs, isPyData, jsFncVal)
    self.__jsFncs.extend(jsIterable)
    return self

  def toStr(self):
    if self.__next is None:
      raise Exception("next() function must be used to avoid infinite loops !!")

    self.__jsFncs.append(self.__next)
    strData = "%s{%s}" % (";".join(self.__while), ";".join(self.__jsFncs))

    self.__jsFncs = []
    return strData





from epyk.core.data import DataClass


class RowGroup(DataClass):

  def activate(self):
    self.dataSrc = 'group'
    return self

  @property
  def className(self):
    """
    Description:
    -----------
    Set the class name to be used for the grouping rows.

    Related Pages:

			https://datatables.net/reference/option/rowGroup.className
    """
    return self._attrs["className"]

  @className.setter
  def className(self, val):
    self._attrs["className"] = val

  @property
  def dataSrc(self):
    """
    Description:
    -----------
    Set the data point to use as the grouping data source

    Related Pages:

			https://datatables.net/reference/option/rowGroup.className
    """
    return self._attrs["dataSrc"]

  @dataSrc.setter
  def dataSrc(self, val):
    self._attrs["dataSrc"] = val

  @property
  def emptyDataGroup(self):
    """
    Description:
    -----------
    Text to show for rows which have null, undefined or empty string group data.

    Related Pages:

			https://datatables.net/reference/option/rowGroup.emptyDataGroup
    """
    return self._attrs["emptyDataGroup"]

  @emptyDataGroup.setter
  def emptyDataGroup(self, val):
    self._attrs["emptyDataGroup"] = val

  @property
  def enable(self):
    """
    Description:
    -----------
    Provides the ability to disable row grouping at initialisation.

    Related Pages:

			https://datatables.net/reference/option/rowGroup.enable
    """
    return self._attrs["enable"]

  @enable.setter
  def enable(self, val):
    self._attrs["enable"] = val

  @property
  def endClassName(self):
    """
    Description:
    -----------
    Set the class name to be used for the grouping end rows.

    Related Pages:

			https://datatables.net/reference/option/rowGroup.endClassName
    """
    return self._attrs["endClassName"]

  @endClassName.setter
  def endClassName(self, val):
    self._attrs["endClassName"] = val

  @property
  def startClassName(self):
    """
    Description:
    -----------
    Set the class name to be used for the grouping start rows.

    Related Pages:

			https://datatables.net/reference/option/rowGroup.startClassName
    """
    return self._attrs["startClassName"]

  @startClassName.setter
  def startClassName(self, val):
    self._attrs["startClassName"] = val
